@@ -1,6 +1,7 @@
 /*
  * Linux /proc/meminfo metrics cluster
  *
+ * Copyright (c) 2013-2016 Red Hat.
  * Copyright (c) 2002 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -13,15 +14,9 @@
  * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  * for more details.
  */
-
-#include <sys/types.h>
-#include <sys/stat.h>
 #include <ctype.h>
-#include <string.h>
-#include <fcntl.h>
-#include <stdio.h>
-
-#include "pmapi.h"
+#include <sys/stat.h>
+#include "linux.h"
 #include "proc_meminfo.h"
 
 static proc_meminfo_t moff;
@@ -32,6 +27,7 @@ static struct {
 } meminfo_fields[] = {
     { "MemTotal",	&moff.MemTotal },
     { "MemFree",	&moff.MemFree },
+    { "MemAvailable",	&moff.MemAvailable },
     { "MemShared",	&moff.MemShared },
     { "Buffers",	&moff.Buffers },
     { "Cached",		&moff.Cached },
@@ -79,8 +75,10 @@ static struct {
     { "HugePages_Free",	&moff.HugepagesFree },
     { "HugePages_Rsvd",	&moff.HugepagesRsvd },
     { "HugePages_Surp",	&moff.HugepagesSurp },
+    { "Hugepagesize",	&moff.Hugepagesize },
     { "DirectMap4k",	&moff.directMap4k },
     { "DirectMap2M",	&moff.directMap2M },
+    { "DirectMap1G",	&moff.directMap1G },
     { NULL, NULL }
 };
 
@@ -90,24 +88,18 @@ static struct {
 int
 refresh_proc_meminfo(proc_meminfo_t *proc_meminfo)
 {
-    static int	started;
     char	buf[1024];
     char	*bufp;
     int64_t	*p;
     int		i;
     FILE	*fp;
 
-    if (!started) {
-	started = 1;
-	memset(proc_meminfo, 0, sizeof(*proc_meminfo));
-    }
-
-    for (i=0; meminfo_fields[i].field != NULL; i++) {
+    for (i = 0; meminfo_fields[i].field != NULL; i++) {
 	p = MOFFSET(i, proc_meminfo);
 	*p = -1; /* marked as "no value available" */
     }
 
-    if ((fp = fopen("/proc/meminfo", "r")) == (FILE *)0)
+    if ((fp = linux_statsfile("/proc/meminfo", buf, sizeof(buf))) == NULL)
 	return -oserror();
 
     while (fgets(buf, sizeof(buf), fp) != NULL) {
@@ -121,7 +113,6 @@ refresh_proc_meminfo(proc_meminfo_t *proc_meminfo)
 	    for (bufp++; *bufp; bufp++) {
 	    	if (isdigit((int)*bufp)) {
 		    sscanf(bufp, "%llu", (unsigned long long *)p);
-		    *p *= 1024; /* kbytes -> bytes */
 		    break;
 		}
 	    }
@@ -130,6 +121,63 @@ refresh_proc_meminfo(proc_meminfo_t *proc_meminfo)
 
     fclose(fp);
 
+    /*
+     * MemAvailable is only in 3.x or later kernels but we can calculate it
+     * using other values, similar to upstream kernel commit 34e431b0ae.
+     */
+    if (!MEMINFO_VALID_VALUE(proc_meminfo->MemAvailable) ||
+	(linux_test_mode & LINUX_TEST_MEMINFO)) {
+	if (MEMINFO_VALID_VALUE(proc_meminfo->MemTotal) &&
+	    MEMINFO_VALID_VALUE(proc_meminfo->MemFree) &&
+	    MEMINFO_VALID_VALUE(proc_meminfo->Active_file) &&
+	    MEMINFO_VALID_VALUE(proc_meminfo->Inactive_file) &&
+	    MEMINFO_VALID_VALUE(proc_meminfo->SlabReclaimable)) {
+
+	    int64_t pagecache;
+	    int64_t wmark_low = 0;
+
+	    /*
+	     * sum for each zone->watermark[WMARK_LOW];
+	     */
+	    if ((fp = linux_statsfile("/proc/zoneinfo", buf, sizeof(buf))) != NULL) {
+		while (fgets(buf, sizeof(buf), fp) != NULL) {
+		    if ((bufp = strstr(buf, "low ")) != NULL) {
+			int64_t low;
+		    	if (sscanf(bufp+4, "%lld", (long long int *)&low) == 1)
+			    wmark_low += low;
+		    }
+		}
+		fclose(fp);
+		wmark_low <<= _pm_pageshift;
+	    }
+
+	    /*  
+	     * Free memory cannot be taken below the low watermark, before the
+	     * system starts swapping.
+	     */
+	    proc_meminfo->MemAvailable = proc_meminfo->MemFree - wmark_low;
+
+	    /*
+	     * Not all the page cache can be freed, otherwise the system will
+	     * start swapping. Assume at least half of the page cache, or the
+	     * low watermark worth of cache, needs to stay.
+	     */
+	    pagecache = proc_meminfo->Active_file + proc_meminfo->Inactive_file;
+	    pagecache -= MIN(pagecache / 2, wmark_low);
+	    proc_meminfo->MemAvailable += pagecache;
+
+	    /*
+	     * Part of the reclaimable slab consists of items that are in use,
+	     * and cannot be freed. Cap this estimate at the low watermark.
+	     */
+	    proc_meminfo->MemAvailable += proc_meminfo->SlabReclaimable;
+	    proc_meminfo->MemAvailable -= MIN(proc_meminfo->SlabReclaimable / 2, wmark_low);
+
+	    if (proc_meminfo->MemAvailable < 0)
+	    	proc_meminfo->MemAvailable = 0;
+	}
+    }
+
     /* success */
     return 0;
 }
