@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2012-2018 Red Hat.
+ * Copyright (c) 2012-2019 Red Hat.
  * Copyright (c) 2009-2010 Aconex. All Rights Reserved.
  * Copyright (c) 1995-2000,2009 Silicon Graphics, Inc. All Rights Reserved.
  *
@@ -36,6 +36,14 @@
 
 static int isDSO = 1;
 static char *username;
+static const char *mmv_prefix = "mmv";
+static const char *pmproxy_prefix = "pmproxy";
+static void *privdata;
+
+static int setup;
+static float fNaN;
+static double dNaN;
+static pmAtomValue aNaN;
 
 /* command line option handling - both short and long options */
 static pmLongOptions longopts[] = {
@@ -52,44 +60,44 @@ static pmdaOptions opts = {
     .long_options = longopts,
 };
 
-static pmdaMetric * metrics;
-static int mtot;
-static pmdaIndom * indoms;
-static int intot;
-
-static pmdaNameSpace * pmns;
-static int reload;			/* require reload of maps */
-static int notify;			/* notify pmcd of changes */
-static int statsdir_code;		/* last statsdir stat code */
-static time_t statsdir_ts;		/* last statsdir timestamp */
-static char * prefix = "mmv";
-
-static char * pcptmpdir;		/* probably /var/tmp */
-static char * pcpvardir;		/* probably /var/pcp */
-static char * pcppmdasdir;		/* probably /var/pcp/pmdas */
-static char pmnsdir[MAXPATHLEN];	/* pcpvardir/pmns */
-static char statsdir[MAXPATHLEN];	/* pcptmpdir/<prefix> */
-
 typedef struct {
-    char *	name;			/* strdup client name */
-    void *	addr;			/* mmap */
-    mmv_disk_value_t * values;		/* values in mmap */
-    mmv_disk_metric_t * metrics1;	/* v1 metric descs in mmap */
-    mmv_disk_metric2_t * metrics2;	/* v2 metric descs in mmap */
-    mmv_disk_label_t * labels; 		/* labels desc in mmap */
-    int		vcnt;			/* number of values */
-    int		mcnt1;			/* number of metrics */
-    int		mcnt2;			/* number of v2 metrics */
-    int		lcnt;			/* number of labels */
-    int		version;		/* v1/v2/v3 version number */
-    int		cluster;		/* cluster identifier */
-    pid_t	pid;			/* process identifier */
-    __int64_t	len;			/* mmap region len */
-    __uint64_t	gen;			/* generation number on open */
+    char		*name;		/* strdup client name */
+    void		*addr;		/* mmap */
+    mmv_disk_value_t	*values;	/* values in mmap */
+    mmv_disk_metric_t	*metrics1;	/* v1 metric descs in mmap */
+    mmv_disk_metric2_t	*metrics2;	/* v2 metric descs in mmap */
+    mmv_disk_label_t	*labels; 	/* labels desc in mmap */
+    int			vcnt;		/* number of values */
+    int			mcnt1;		/* number of metrics */
+    int			mcnt2;		/* number of v2 metrics */
+    int			lcnt;		/* number of labels */
+    int			version;	/* v1/v2/v3 version number */
+    int			cluster;	/* cluster identifier */
+    pid_t		pid;		/* process identifier */
+    __int64_t		len;		/* mmap region len */
+    __uint64_t		gen;		/* generation number on open */
 } stats_t;
 
-static stats_t * slist;
-static int scnt;
+typedef struct {
+    pmdaMetric		*metrics;
+    pmdaIndom		*indoms;
+    pmdaNameSpace	*pmns;
+    stats_t		*slist;
+    int			scnt;
+    int			mtot;
+    int			intot;
+    int			reload;		/* require reload of maps */
+    int			notify;		/* notify pmcd of changes */
+    int			statsdir_code;	/* last statsdir stat code */
+    time_t		statsdir_ts;	/* last statsdir timestamp */
+    const char		*prefix;
+    char		*pcptmpdir;		/* probably /var/tmp */
+    char		*pcpvardir;		/* probably /var/pcp */
+    char		*pcppmdasdir;		/* probably /var/pcp/pmdas */
+    char		pmnsdir[MAXPATHLEN];	/* pcpvardir/pmns */
+    char		statsdir[MAXPATHLEN];	/* pcptmpdir/<prefix> */
+    char		buffer[MMV_STRINGMAX];	/* temporary fetch buffer */
+} agent_t;
 
 #define MAX_MMV_COUNT 10000		/* enforce reasonable limits */
 #define MAX_MMV_CLUSTER ((1<<12)-1)
@@ -108,15 +116,15 @@ valid_cluster(int requested)
  * If a specific (non-zero) cluster is requested we always use it.
  */
 static int
-choose_cluster(int requested, const char *path)
+choose_cluster(agent_t *ap, int requested, const char *path)
 {
     int i;
 
     if (!requested) {
 	int next_cluster = 1;
 
-	for (i = 0; i < scnt; i++) {
-	    if (slist[i].cluster == next_cluster) {
+	for (i = 0; i < ap->scnt; i++) {
+	    if (ap->slist[i].cluster == next_cluster) {
 		next_cluster++;
 		i = 0;	/* restart, we're filling holes */
 	    }
@@ -130,8 +138,8 @@ choose_cluster(int requested, const char *path)
     if (!valid_cluster(requested))
 	return -EINVAL;
 
-    for (i = 0; i < scnt; i++) {
-	if (slist[i].cluster == requested) {
+    for (i = 0; i < ap->scnt; i++) {
+	if (ap->slist[i].cluster == requested) {
 	    if (pmDebugOptions.appl0)
 		pmNotifyErr(LOG_DEBUG,
 				"MMV: %s: duplicate cluster %d in use",
@@ -143,9 +151,14 @@ choose_cluster(int requested, const char *path)
 }
 
 static int
-create_client_stat(const char *client, const char *path, size_t size)
+create_client_stat(agent_t *ap, const char *client, const char *path, size_t size)
 {
-    int fd;
+    mmv_disk_header_t	header;
+    stats_t		*sp;
+    size_t		offset;
+    void		*m;
+    int			cluster;
+    int			in, fd;
 
     if (pmDebugOptions.appl0)
 	pmNotifyErr(LOG_DEBUG, "MMV: create_client_stat: %s, %s", client, path);
@@ -155,14 +168,11 @@ create_client_stat(const char *client, const char *path, size_t size)
 	return -EINVAL;
 
     if ((fd = open(path, O_RDONLY)) >= 0) {
-	void *m = __pmMemoryMap(fd, size, 0);
-
+	m = __pmMemoryMap(fd, size, 0);
 	close(fd);
-	if (m != NULL) {
-	    mmv_disk_header_t header = *(mmv_disk_header_t *)m;
-	    size_t offset;
-	    int cluster;
 
+	if (m != NULL) {
+	    header = *(mmv_disk_header_t *)m;
 	    if (strncmp(header.magic, "MMV", 4)) {
 		__pmMemoryUnmap(m, size);
 		return -EINVAL;
@@ -174,7 +184,7 @@ create_client_stat(const char *client, const char *path, size_t size)
 		if (pmDebugOptions.appl0)
 		    pmNotifyErr(LOG_ERR,
 			"%s: %s client version %d unsupported (current is %d)",
-			pmGetProgname(), prefix, header.version, MMV_VERSION);
+			pmGetProgname(), ap->prefix, header.version, MMV_VERSION);
 		__pmMemoryUnmap(m, size);
 		return -ENOSYS;
 	    }
@@ -203,7 +213,7 @@ create_client_stat(const char *client, const char *path, size_t size)
 		return -ESRCH;
 	    }
 
-	    if ((cluster = choose_cluster(header.cluster, path)) < 0) {
+	    if ((cluster = choose_cluster(ap, header.cluster, path)) < 0) {
 		__pmMemoryUnmap(m, size);
 		return cluster;
 	    }
@@ -211,25 +221,29 @@ create_client_stat(const char *client, const char *path, size_t size)
 	    /* all checks out so far, we'll use this one */
 	    if (pmDebugOptions.appl0)
 		pmNotifyErr(LOG_DEBUG, "MMV: loading %s client: %d \"%s\"",
-				    prefix, cluster, path);
+				    ap->prefix, cluster, path);
 
-	    slist = realloc(slist, sizeof(stats_t) * (scnt + 1));
-	    if (slist != NULL) {
-		memset(&slist[scnt], 0, sizeof(stats_t));
-		slist[scnt].name = strdup(client);
-		slist[scnt].addr = m;
+	    in = ap->scnt;
+	    sp = realloc(ap->slist, sizeof(stats_t) * (in + 1));
+	    if (sp != NULL) {
+		memset(&sp[in], 0, sizeof(stats_t));
+		sp[in].name = strdup(client);
+		sp[in].addr = m;
 		if ((header.flags & MMV_FLAG_PROCESS) != 0)
-		    slist[scnt].pid = (pid_t)header.process;
-		slist[scnt].version = header.version;
-		slist[scnt].cluster = cluster;
-		slist[scnt].gen = header.g1;
-		slist[scnt].len = size;
-		scnt++;
+		    sp[in].pid = (pid_t)header.process;
+		sp[in].version = header.version;
+		sp[in].cluster = cluster;
+		sp[in].gen = header.g1;
+		sp[in].len = size;
+		ap->slist = sp;
+		ap->scnt++;
 	    } else {
 		pmNotifyErr(LOG_ERR, "%s: client \"%s\" out of memory - %s",
 				pmGetProgname(), client, osstrerror());
 		__pmMemoryUnmap(m, size);
-		scnt = 0;
+		ap->scnt = 0;
+		free(ap->slist);
+		ap->slist = NULL;
 	    }
 	} else {
 	    pmNotifyErr(LOG_ERR, "%s: failed to memory map \"%s\" - %s",
@@ -244,10 +258,10 @@ create_client_stat(const char *client, const char *path, size_t size)
 
 /* check validity of client metric name, return non-zero if bad or duplicate */
 static int
-verify_metric_name(const char *name, int pos, stats_t *s)
+verify_metric_name(agent_t *ap, const char *name, int pos, stats_t *s)
 {
-    const char *p = name;
-    pmID pmid;
+    const char		*p = name;
+    pmID		pmid;
 
     if (pmDebugOptions.appl0)
 	pmNotifyErr(LOG_DEBUG, "MMV: verify_metric_name: %s", name);
@@ -264,7 +278,7 @@ verify_metric_name(const char *name, int pos, stats_t *s)
 			    pos, s->name, *p);
 	return -EINVAL;
     }
-    if (pmdaTreePMID(pmns, name, &pmid) == 0)
+    if (pmdaTreePMID(ap->pmns, name, &pmid) == 0)
 	return -EEXIST;
     return 0;
 }
@@ -288,44 +302,47 @@ static int
 create_metric(pmdaExt *pmda, stats_t *s, char *name, pmID pmid, unsigned indom,
 	mmv_metric_type_t type, mmv_metric_sem_t semantics, pmUnits units)
 {
+    pmdaMetric		*mp;
+    agent_t		*ap = (agent_t *)pmdaExtGetData(pmda);
+
     if (pmDebugOptions.appl0)
 	pmNotifyErr(LOG_DEBUG, "MMV: create_metric: %s - %s", name, pmIDStr(pmid));
 
-    metrics = realloc(metrics, sizeof(pmdaMetric) * (mtot + 1));
-    if (metrics == NULL)  {
+    mp = realloc(ap->metrics, sizeof(pmdaMetric) * (ap->mtot + 1));
+    if (mp == NULL)  {
 	pmNotifyErr(LOG_ERR, "cannot grow MMV metric list: %s", s->name);
 	return -ENOMEM;
     }
-
-    metrics[mtot].m_user = NULL;
-    metrics[mtot].m_desc.pmid = pmid;
+    ap->metrics = mp;
+    ap->metrics[ap->mtot].m_user = ap;
+    ap->metrics[ap->mtot].m_desc.pmid = pmid;
 
     if (type == MMV_TYPE_ELAPSED) {
 	pmUnits unit = PMDA_PMUNITS(0,1,0,0,PM_TIME_USEC,0);
-	metrics[mtot].m_desc.sem = PM_SEM_COUNTER;
-	metrics[mtot].m_desc.type = MMV_TYPE_I64;
-	metrics[mtot].m_desc.units = unit;
+	ap->metrics[ap->mtot].m_desc.sem = PM_SEM_COUNTER;
+	ap->metrics[ap->mtot].m_desc.type = MMV_TYPE_I64;
+	ap->metrics[ap->mtot].m_desc.units = unit;
     } else {
 	if (semantics)
-	    metrics[mtot].m_desc.sem = semantics;
+	    ap->metrics[ap->mtot].m_desc.sem = semantics;
 	else
-	    metrics[mtot].m_desc.sem = PM_SEM_COUNTER;
-	metrics[mtot].m_desc.type = type;
-	memcpy(&metrics[mtot].m_desc.units, &units, sizeof(pmUnits));
+	    ap->metrics[ap->mtot].m_desc.sem = PM_SEM_COUNTER;
+	ap->metrics[ap->mtot].m_desc.type = type;
+	memcpy(&ap->metrics[ap->mtot].m_desc.units, &units, sizeof(pmUnits));
     }
     if (!indom || indom == PM_INDOM_NULL)
-	metrics[mtot].m_desc.indom = PM_INDOM_NULL;
+	ap->metrics[ap->mtot].m_desc.indom = PM_INDOM_NULL;
     else
-	metrics[mtot].m_desc.indom = 
+	ap->metrics[ap->mtot].m_desc.indom = 
 		pmInDom_build(pmda->e_domain, (s->cluster << 11) | indom);
 
     if (pmDebugOptions.appl0)
 	pmNotifyErr(LOG_DEBUG,
 			"MMV: map_stats adding metric[%d] %s %s from %s\n",
-			mtot, name, pmIDStr(pmid), s->name);
+			ap->mtot, name, pmIDStr(pmid), s->name);
 
-    mtot++;
-    pmdaTreeInsert(pmns, pmid, name);
+    ap->mtot++;
+    pmdaTreeInsert(ap->pmns, pmid, name);
 
     return 0;
 }
@@ -334,7 +351,8 @@ create_metric(pmdaExt *pmda, stats_t *s, char *name, pmID pmid, unsigned indom,
 static int
 verify_indom_serial(pmdaExt *pmda, int serial, stats_t *s, pmInDom *p, pmdaIndom **i)
 {
-    int index;
+    agent_t		*ap = (agent_t *)pmdaExtGetData(pmda);
+    int			index;
 
     if (pmDebugOptions.appl0)
 	pmNotifyErr(LOG_DEBUG, "MMV: verify_indom_serial: %u", serial);
@@ -346,9 +364,9 @@ verify_indom_serial(pmdaExt *pmda, int serial, stats_t *s, pmInDom *p, pmdaIndom
     }
 
     *p = pmInDom_build(pmda->e_domain, (s->cluster << 11) | serial);
-    for (index = 0; index < intot; index++) {
-	*i = &indoms[index];
-	if (indoms[index].it_indom == *p)
+    for (index = 0; index < ap->intot; index++) {
+	*i = &ap->indoms[index];
+	if (ap->indoms[index].it_indom == *p)
 	    return -EEXIST;
     }
     *i = NULL;
@@ -361,8 +379,9 @@ update_indom(pmdaExt *pmda, stats_t *s, __uint64_t offset, __uint32_t count,
 {
     mmv_disk_instance_t *in1 = NULL;
     mmv_disk_instance2_t *in2 = NULL;
-    mmv_disk_string_t *string;
-    int i, j, size, newinsts = 0;
+    mmv_disk_string_t	*string;
+    pmdaInstid		*iip;
+    int			i, j, size, newinsts = 0;
 
     if (pmDebugOptions.appl0)
 	pmNotifyErr(LOG_DEBUG, "MMV: update_indom on %s: %u (%d insts, count %d)",
@@ -398,13 +417,16 @@ update_indom(pmdaExt *pmda, stats_t *s, __uint64_t offset, __uint32_t count,
 
     /* allocate memory, then append new instances to the known set */
     size = sizeof(pmdaInstid) * (ip->it_numinst + newinsts);
-    ip->it_set = (pmdaInstid *)realloc(ip->it_set, size);
-    if (ip->it_set == NULL) {
+    iip = (pmdaInstid *)realloc(ip->it_set, size);
+    if (iip == NULL) {
 	pmNotifyErr(LOG_ERR, "%s: cannot get memory for instance list in %s",
 			pmGetProgname(), s->name);
 	ip->it_numinst = 0;
+	free(ip->it_set);
+	ip->it_set = NULL;
 	return -ENOMEM;
     }
+    ip->it_set = iip;
 
     if (s->version == MMV_VERSION1) {
 	for (i = 0; i < count; i++) {
@@ -438,22 +460,24 @@ static int
 create_indom(pmdaExt *pmda, stats_t *s, __uint64_t offset, __uint32_t count,
 		mmv_disk_indom_t *id, pmInDom indom)
 {
-    int i;
-    pmdaIndom *ip;
-    mmv_disk_instance_t *in1;
+    agent_t		*ap = (agent_t *)pmdaExtGetData(pmda);
+    pmdaIndom		*ip;
+    mmv_disk_instance_t	*in1;
     mmv_disk_instance2_t *in2;
-    mmv_disk_string_t *string;
+    mmv_disk_string_t	*string;
+    int			i;
 
     if (pmDebugOptions.appl0)
 	pmNotifyErr(LOG_DEBUG, "MMV: create_indom: %u", id->serial);
 
-    indoms = realloc(indoms, sizeof(pmdaIndom) * (intot + 1));
-    if (indoms == NULL) {
+    ip = realloc(ap->indoms, sizeof(pmdaIndom) * (ap->intot + 1));
+    if (ip == NULL) {
 	pmNotifyErr(LOG_ERR, "%s: cannot grow indom list in %s",
 			pmGetProgname(), s->name);
 	return -ENOMEM;
     }
-    ip = &indoms[intot++];
+    ap->indoms = ip;
+    ip = &ap->indoms[ap->intot++];
     ip->it_indom = indom;
     ip->it_set = (pmdaInstid *)calloc(count, sizeof(pmdaInstid));
     if (ip->it_set == NULL) {
@@ -486,65 +510,62 @@ create_indom(pmdaExt *pmda, stats_t *s, __uint64_t offset, __uint32_t count,
 static void
 map_stats(pmdaExt *pmda)
 {
-    struct dirent **files;
-    char name[64];
-    int need_reload = 0;
-    int i, j, k, sts, num;
-    int sep = pmPathSeparator();
+    struct dirent	**files;
+    struct stat		statbuf;
+    agent_t		*ap = (agent_t *)pmdaExtGetData(pmda);
+    char		path[MAXPATHLEN], name[64], *client;
+    int			need_reload = 0, sep = pmPathSeparator();
+    int			i, j, k, sts, num;
 
-    if (pmns) {
-	pmdaTreeRelease(pmns);
-	notify |= PMDA_EXT_NAMES_CHANGE;
+    if (ap->pmns) {
+	pmdaTreeRelease(ap->pmns);
+	ap->notify |= PMDA_EXT_NAMES_CHANGE;
     }
 
-    if ((sts = pmdaTreeCreate(&pmns)) < 0) {
+    if ((sts = pmdaTreeCreate(&ap->pmns)) < 0) {
 	pmNotifyErr(LOG_ERR, "%s: failed to create new pmns: %s\n",
 			pmGetProgname(), pmErrStr(sts));
-	pmns = NULL;
+	ap->pmns = NULL;
 	return;
     }
 
     /* hard-coded metrics (not from mmap'd files) */
-    pmsprintf(name, sizeof(name), "%s.control.reload", prefix);
-    pmdaTreeInsert(pmns, pmID_build(pmda->e_domain, 0, 0), name);
-    pmsprintf(name, sizeof(name), "%s.control.debug", prefix);
-    pmdaTreeInsert(pmns, pmID_build(pmda->e_domain, 0, 1), name);
-    pmsprintf(name, sizeof(name), "%s.control.files", prefix);
-    pmdaTreeInsert(pmns, pmID_build(pmda->e_domain, 0, 2), name);
-    mtot = 3;
+    pmsprintf(name, sizeof(name), "%s.control.reload", ap->prefix);
+    pmdaTreeInsert(ap->pmns, pmID_build(pmda->e_domain, 0, 0), name);
+    pmsprintf(name, sizeof(name), "%s.control.debug", ap->prefix);
+    pmdaTreeInsert(ap->pmns, pmID_build(pmda->e_domain, 0, 1), name);
+    pmsprintf(name, sizeof(name), "%s.control.files", ap->prefix);
+    pmdaTreeInsert(ap->pmns, pmID_build(pmda->e_domain, 0, 2), name);
+    ap->mtot = 3;
 
-    if (indoms != NULL) {
-	for (i = 0; i < intot; i++)
-	    free(indoms[i].it_set);
-	free(indoms);
-	indoms = NULL;
-	intot = 0;
+    if (ap->indoms != NULL) {
+	for (i = 0; i < ap->intot; i++)
+	    free(ap->indoms[i].it_set);
+	free(ap->indoms);
+	ap->indoms = NULL;
+	ap->intot = 0;
     }
 
-    if (slist != NULL) {
-	for (i = 0; i < scnt; i++) {
-	    free(slist[i].name);
-	    __pmMemoryUnmap(slist[i].addr, slist[i].len);
+    if (ap->slist != NULL) {
+	for (i = 0; i < ap->scnt; i++) {
+	    free(ap->slist[i].name);
+	    __pmMemoryUnmap(ap->slist[i].addr, ap->slist[i].len);
 	}
-	free(slist);
-	slist = NULL;
-	scnt = 0;
+	free(ap->slist);
+	ap->slist = NULL;
+	ap->scnt = 0;
     }
 
-    num = scandir(statsdir, &files, NULL, NULL);
+    num = scandir(ap->statsdir, &files, NULL, alphasort);
     for (i = 0; i < num; i++) {
-	struct stat statbuf;
-	char path[MAXPATHLEN];
-	char *client;
-
 	if (files[i]->d_name[0] == '.')
 	    continue;
 
 	client = files[i]->d_name;
-	pmsprintf(path, sizeof(path), "%s%c%s", statsdir, sep, client);
+	pmsprintf(path, sizeof(path), "%s%c%s", ap->statsdir, sep, client);
 
 	if (stat(path, &statbuf) >= 0 && S_ISREG(statbuf.st_mode))
-	    if (create_client_stat(client, path, statbuf.st_size) == -EAGAIN)
+	    if (create_client_stat(ap, client, path, statbuf.st_size) == -EAGAIN)
 		need_reload = 1;
     }
 
@@ -553,8 +574,8 @@ map_stats(pmdaExt *pmda)
     if (num > 0)
 	free(files);
 
-    for (i = 0; slist && i < scnt; i++) {
-	stats_t *s = slist + i;
+    for (i = 0; ap->slist && i < ap->scnt; i++) {
+	stats_t	*s = ap->slist + i;
 	mmv_disk_indom_t *id;
 	mmv_disk_header_t *hdr = (mmv_disk_header_t *)s->addr;
 	mmv_disk_toc_t *toc = (mmv_disk_toc_t *)
@@ -599,11 +620,11 @@ map_stats(pmdaExt *pmda)
 
 			/* build name, check its legitimate and unique */
 			if (hdr->flags & MMV_FLAG_NOPREFIX)
-			    pmsprintf(name, sizeof(name), "%s.", prefix);
+			    pmsprintf(name, sizeof(name), "%s.", ap->prefix);
 			else
-			    pmsprintf(name, sizeof(name), "%s.%s.", prefix, s->name);
+			    pmsprintf(name, sizeof(name), "%s.%s.", ap->prefix, s->name);
 			strcat(name, mp->name);
-			if (verify_metric_name(name, k, s) != 0)
+			if (verify_metric_name(ap, name, k, s) != 0)
 			    continue;
 			if (verify_metric_item(mp->item, name, s) != 0)
 			    continue;
@@ -653,12 +674,12 @@ map_stats(pmdaExt *pmda)
 
 			/* build name, check its legitimate and unique */
 			if (hdr->flags & MMV_FLAG_NOPREFIX)
-			    pmsprintf(name, sizeof(name), "%s.", prefix);
+			    pmsprintf(name, sizeof(name), "%s.", ap->prefix);
 			else
-			    pmsprintf(name, sizeof(name), "%s.%s.", prefix, s->name);
+			    pmsprintf(name, sizeof(name), "%s.%s.", ap->prefix, s->name);
 			strcat(name, buf);
 
-			if (verify_metric_name(name, k, s) != 0)
+			if (verify_metric_name(ap, name, k, s) != 0)
 			    continue;
 			if (verify_metric_item(mp->item, name, s) != 0)
 			    continue;
@@ -805,8 +826,8 @@ map_stats(pmdaExt *pmda)
 	}
     }
 
-    pmdaTreeRebuildHash(pmns, mtot);	/* for reverse (pmid->name) lookups */
-    reload = need_reload;
+    pmdaTreeRebuildHash(ap->pmns, ap->mtot); /* for reverse (pmid->name) lookups */
+    ap->reload = need_reload;
 }
 
 static int
@@ -814,9 +835,9 @@ mmv_lookup_item1(int item, unsigned int inst,
 	stats_t *s, mmv_disk_value_t **value,
 	__uint64_t *shorttext, __uint64_t *helptext)
 {
-    mmv_disk_metric_t *m1 = s->metrics1;
-    mmv_disk_value_t *v = s->values;
-    int mi, vi, sts = PM_ERR_PMID;
+    mmv_disk_metric_t	*m1 = s->metrics1;
+    mmv_disk_value_t	*v = s->values;
+    int			mi, vi, sts = PM_ERR_PMID;
 
     for (mi = 0; mi < s->mcnt1; mi++) {
 	if (m1[mi].item != item)
@@ -849,9 +870,9 @@ mmv_lookup_item2(int item, unsigned int inst,
 	stats_t *s, mmv_disk_value_t **value,
 	__uint64_t *shorttext, __uint64_t *helptext)
 {
-    mmv_disk_metric2_t *m2 = s->metrics2;
-    mmv_disk_value_t *v = s->values;
-    int mi, vi, sts = PM_ERR_PMID;
+    mmv_disk_metric2_t	*m2 = s->metrics2;
+    mmv_disk_value_t	*v = s->values;
+    int			mi, vi, sts = PM_ERR_PMID;
 
     for (mi = 0; mi < s->mcnt2; mi++) {
 	if (m2[mi].item != item)
@@ -880,14 +901,14 @@ mmv_lookup_item2(int item, unsigned int inst,
 }
 
 static int
-mmv_lookup_stat_metric(pmID pmid, unsigned int inst,
+mmv_lookup_stat_metric(agent_t *agent, pmID pmid, unsigned int inst,
 	stats_t **stats, mmv_disk_value_t **value,
 	__uint64_t *shorttext, __uint64_t *helptext)
 {
-    int si, sts = PM_ERR_PMID;
+    int			si, sts = PM_ERR_PMID;
 
-    for (si = 0; si < scnt; si++) {
-	stats_t *s = &slist[si];
+    for (si = 0; si < agent->scnt; si++) {
+	stats_t *s = &agent->slist[si];
 
 	if (s->cluster != pmID_cluster(pmid))
 	    continue;
@@ -895,6 +916,8 @@ mmv_lookup_stat_metric(pmID pmid, unsigned int inst,
 	sts = (s->version == MMV_VERSION1) ?
 	    mmv_lookup_item1(pmID_item(pmid), inst, s, value, shorttext, helptext):
 	    mmv_lookup_item2(pmID_item(pmid), inst, s, value, shorttext, helptext);
+	if (sts == MMV_TYPE_NOSUPPORT)
+	    sts = PM_ERR_APPVERSION;
 	if (sts >= 0) {
 	    *stats = s;
 	    break;
@@ -904,10 +927,10 @@ mmv_lookup_stat_metric(pmID pmid, unsigned int inst,
 }
 
 static int
-mmv_lookup_stat_metric_value(pmID pmid, unsigned int inst,
+mmv_lookup_stat_metric_value(agent_t *agent, pmID pmid, unsigned int inst,
 	stats_t **stats, mmv_disk_value_t **value)
 {
-    return mmv_lookup_stat_metric(pmid, inst, stats, value, NULL, NULL);
+    return mmv_lookup_stat_metric(agent, pmid, inst, stats, value, NULL, NULL);
 }
 
 /*
@@ -916,60 +939,59 @@ mmv_lookup_stat_metric_value(pmID pmid, unsigned int inst,
 static int
 mmv_fetchCallBack(pmdaMetric *mdesc, unsigned int inst, pmAtomValue *atom)
 {
-    if (pmID_cluster(mdesc->m_desc.pmid) == 0) {
-	if (pmID_item(mdesc->m_desc.pmid) <= 2) {
-	    atom->l = *(int *)mdesc->m_user;
-	    return PMDA_FETCH_STATIC;
+    mmv_disk_string_t	*str;
+    mmv_disk_value_t	*v;
+    __uint64_t		offset;
+    agent_t		*ap = (agent_t *)mdesc->m_user;
+    stats_t		*s;
+    pmID		pmid = mdesc->m_desc.pmid;
+    int			sts, flags;
+
+    if (pmID_cluster(pmid) == 0) {
+	switch(pmID_item(pmid)) {
+	    case 0:
+		atom->l = ap->reload;
+		return PMDA_FETCH_STATIC;
+	    case 1:
+		atom->l = pmDebug;
+		return PMDA_FETCH_STATIC;
+	    case 2:
+		atom->l = ap->scnt;
+		return PMDA_FETCH_STATIC;
+	    default:
+		break;
 	}
 	return PM_ERR_PMID;
+    }
 
-    } else if (scnt > 0) {	/* We have at least one source of metrics */
-	static int setup;
-	static float fNaN;
-	static double dNaN;
-	static pmAtomValue aNaN;
-	static char buffer[MMV_STRINGMAX];
-	mmv_disk_string_t *str;
-	mmv_disk_value_t *v;
-	__uint64_t offset;
-	stats_t *s;
-	int rv, fl;
+    if (ap->scnt > 0) {	/* We have at least one source of metrics */
+	if ((sts = mmv_lookup_stat_metric_value(ap, pmid, inst, &s, &v)) < 0)
+	    return sts;
+	flags = ((mmv_disk_header_t *)s->addr)->flags;
 
-	rv = mmv_lookup_stat_metric_value(mdesc->m_desc.pmid, inst, &s, &v);
-	if (rv < 0)
-	    return rv;
-	fl = ((mmv_disk_header_t *)s->addr)->flags;
-
-	if (!setup) {
-	    setup++;
-	    fNaN = (float)0.0 / (float)0.0;
-	    dNaN = (double)0.0 / (double)0.0;
-	    memset(&aNaN, -1, sizeof(pmAtomValue));
-	}
-
-	switch (rv) {
+	switch (sts) {
 	    case MMV_TYPE_I32:
 	    case MMV_TYPE_U32:
 	    case MMV_TYPE_I64:
 	    case MMV_TYPE_U64:
 		memcpy(atom, &v->value, sizeof(pmAtomValue));
-		if ((fl & MMV_FLAG_SENTINEL) &&
+		if ((flags & MMV_FLAG_SENTINEL) &&
 		    (memcmp(atom, &aNaN, sizeof(*atom)) == 0))
 		    return PMDA_FETCH_NOVALUES;
 		break;
 	    case MMV_TYPE_FLOAT:
 		memcpy(atom, &v->value, sizeof(pmAtomValue));
-		if ((fl & MMV_FLAG_SENTINEL) && atom->f == fNaN)
+		if ((flags & MMV_FLAG_SENTINEL) && atom->f == fNaN)
 		    return PMDA_FETCH_NOVALUES;
 		break;
 	    case MMV_TYPE_DOUBLE:
 		memcpy(atom, &v->value, sizeof(pmAtomValue));
-		if ((fl & MMV_FLAG_SENTINEL) && atom->d == dNaN)
+		if ((flags & MMV_FLAG_SENTINEL) && atom->d == dNaN)
 		    return PMDA_FETCH_NOVALUES;
 		break;
 	    case MMV_TYPE_ELAPSED: {
 		atom->ll = v->value.ll;
-		if ((fl & MMV_FLAG_SENTINEL) &&
+		if ((flags & MMV_FLAG_SENTINEL) &&
 		    (memcmp(atom, &aNaN, sizeof(*atom)) == 0))
 		    return 0;
 		if (v->extra < 0) {	/* inside a timed section */
@@ -990,15 +1012,13 @@ mmv_fetchCallBack(pmdaMetric *mdesc, unsigned int inst, pmAtomValue *atom)
 		    return PM_ERR_GENERIC;
 		}
 		str = (mmv_disk_string_t *)((char *)s->addr + offset);
-		memcpy(buffer, str->payload, sizeof(buffer));
-		if ((fl & MMV_FLAG_SENTINEL) && buffer[0] == '\0')
+		memcpy(ap->buffer, str->payload, sizeof(ap->buffer));
+		if ((flags & MMV_FLAG_SENTINEL) && ap->buffer[0] == '\0')
 		    return PMDA_FETCH_NOVALUES;
-		buffer[sizeof(buffer)-1] = '\0';
-		atom->cp = buffer;
+		ap->buffer[sizeof(ap->buffer)-1] = '\0';
+		atom->cp = ap->buffer;
 		break;
 	    }
-	    case MMV_TYPE_NOSUPPORT:
-		return PM_ERR_APPVERSION;
 	}
 	return PMDA_FETCH_STATIC;
     }
@@ -1009,18 +1029,18 @@ mmv_fetchCallBack(pmdaMetric *mdesc, unsigned int inst, pmAtomValue *atom)
 static void
 mmv_reload_maybe(pmdaExt *pmda)
 {
-    int i;
-    struct stat s;
-    int need_reload = reload;
+    struct stat		s;
+    agent_t		*ap = (agent_t *)pmdaExtGetData(pmda);
+    int			i, need_reload = ap->reload;
 
     /* check if generation numbers changed or monitored process exited */
-    for (i = 0; i < scnt; i++) {
-	mmv_disk_header_t *hdr = (mmv_disk_header_t *)slist[i].addr;
-	if (hdr->g1 != slist[i].gen || hdr->g2 != slist[i].gen) {
+    for (i = 0; i < ap->scnt; i++) {
+	mmv_disk_header_t *hdr = (mmv_disk_header_t *)ap->slist[i].addr;
+	if (hdr->g1 != ap->slist[i].gen || hdr->g2 != ap->slist[i].gen) {
 	    need_reload++;
 	    break;
 	}
-	if (slist[i].pid && !__pmProcessExists(slist[i].pid)) {
+	if (ap->slist[i].pid && !__pmProcessExists(ap->slist[i].pid)) {
 	    need_reload++;
 	    break;
 	}
@@ -1032,17 +1052,17 @@ mmv_reload_maybe(pmdaExt *pmda)
      * a change in permissions from accessible to not (or vice-
      * versa), and so on.
      */
-    if (stat(statsdir, &s) >= 0) {
-	if (s.st_mtime != statsdir_ts) {
+    if (stat(ap->statsdir, &s) >= 0) {
+	if (s.st_mtime != ap->statsdir_ts) {
 	    need_reload++;
-	    statsdir_code = 0;
-	    statsdir_ts = s.st_mtime;
+	    ap->statsdir_code = 0;
+	    ap->statsdir_ts = s.st_mtime;
 	}
     } else {
 	i = oserror();
-	if (statsdir_code != i) {
-	    statsdir_code = i;
-	    statsdir_ts = 0;
+	if (ap->statsdir_code != i) {
+	    ap->statsdir_code = i;
+	    ap->statsdir_ts = 0;
 	    need_reload++;
 	}
     }
@@ -1052,36 +1072,35 @@ mmv_reload_maybe(pmdaExt *pmda)
 	    pmNotifyErr(LOG_DEBUG, "MMV: %s: reloading", pmGetProgname());
 	map_stats(pmda);
 
-	pmda->e_indoms = indoms;
-	pmda->e_nindoms = intot;
-	pmdaRehash(pmda, metrics, mtot);
+	pmda->e_indoms = ap->indoms;
+	pmda->e_nindoms = ap->intot;
+	pmdaRehash(pmda, ap->metrics, ap->mtot);
 
 	if (pmDebugOptions.appl0)
 	    pmNotifyErr(LOG_DEBUG, 
 		      "MMV: %s: %d metrics and %d indoms after reload", 
-		      pmGetProgname(), mtot, intot);
+		      pmGetProgname(), ap->mtot, ap->intot);
     }
 }
 
 /* Intercept request for descriptor and check if we'd have to reload */
 static int
-mmv_desc(pmID pmid, pmDesc *desc, pmdaExt *ep)
+mmv_desc(pmID pmid, pmDesc *desc, pmdaExt *pmda)
 {
-    mmv_reload_maybe(ep);
-    return pmdaDesc(pmid, desc, ep);
+    mmv_reload_maybe(pmda);
+    return pmdaDesc(pmid, desc, pmda);
 }
 
 static int
-mmv_lookup_metric_helptext(pmID pmid, int type, char **text)
+mmv_lookup_metric_helptext(agent_t *ap, pmID pmid, int type, char **text)
 {
-    static char string[MMV_STRINGMAX];
-    mmv_disk_string_t *str;
-    mmv_disk_value_t *v;
-    __uint64_t st, lt;
-    size_t offset;
-    stats_t *s;
+    mmv_disk_string_t	*str;
+    mmv_disk_value_t	*v;
+    __uint64_t		st, lt;
+    size_t		offset;
+    stats_t		*s;
 
-    if (mmv_lookup_stat_metric(pmid, PM_IN_NULL, &s, &v, &st, &lt) < 0)
+    if (mmv_lookup_stat_metric(ap, pmid, PM_IN_NULL, &s, &v, &st, &lt) < 0)
 	return PM_ERR_PMID;
 
     if ((type & PM_TEXT_ONELINE) && st) {
@@ -1096,9 +1115,9 @@ mmv_lookup_metric_helptext(pmID pmid, int type, char **text)
 	offset -= sizeof(mmv_disk_string_t);
 
 	str = (mmv_disk_string_t *)((char *)s->addr + offset);
-	memcpy(string, str->payload, sizeof(string));
-	string[sizeof(string)-1] = '\0';
-	*text = string;
+	memcpy(ap->buffer, str->payload, sizeof(ap->buffer));
+	ap->buffer[sizeof(ap->buffer)-1] = '\0';
+	*text = ap->buffer;
 	return 0;
     }
 
@@ -1114,9 +1133,9 @@ mmv_lookup_metric_helptext(pmID pmid, int type, char **text)
 	offset -= sizeof(mmv_disk_string_t);
 
 	str = (mmv_disk_string_t *)((char *)s->addr + offset);
-	memcpy(string, str->payload, sizeof(string));
-	string[sizeof(string)-1] = '\0';
-	*text = string;
+	memcpy(ap->buffer, str->payload, sizeof(ap->buffer));
+	ap->buffer[sizeof(ap->buffer)-1] = '\0';
+	*text = ap->buffer;
 	return 0;
     }
 
@@ -1124,12 +1143,14 @@ mmv_lookup_metric_helptext(pmID pmid, int type, char **text)
 }
 
 static int
-mmv_text(int ident, int type, char **buffer, pmdaExt *ep)
+mmv_text(int ident, int type, char **buffer, pmdaExt *pmda)
 {
+    agent_t		*agent = (agent_t *)pmdaExtGetData(pmda);
+
     if (type & PM_TEXT_INDOM)
 	return PM_ERR_TEXT;
 
-    mmv_reload_maybe(ep);
+    mmv_reload_maybe(pmda);
     if (pmID_cluster(ident) == 0) {
 	switch (pmID_item(ident)) {
 	    case 0: {
@@ -1161,42 +1182,46 @@ mmv_text(int ident, int type, char **buffer, pmdaExt *ep)
 	return PM_ERR_PMID;
     }
 
-    return mmv_lookup_metric_helptext(ident, type, buffer);
+    return mmv_lookup_metric_helptext(agent, ident, type, buffer);
 }
 
 static int
 mmv_instance(pmInDom indom, int inst, char *name, 
-	     pmInResult **result, pmdaExt *ep)
+	     pmInResult **result, pmdaExt *pmda)
 {
-    mmv_reload_maybe(ep);
-    return pmdaInstance(indom, inst, name, result, ep);
+    mmv_reload_maybe(pmda);
+    return pmdaInstance(indom, inst, name, result, pmda);
 }
 
 static int
 mmv_fetch(int numpmid, pmID pmidlist[], pmResult **resp, pmdaExt *pmda)
 {
+    agent_t		*ap = (agent_t *)pmdaExtGetData(pmda);
+
     mmv_reload_maybe(pmda);
-    if (notify) {
-	pmdaExtSetFlags(pmda, notify);
-	notify = 0;
+    if (ap->notify) {
+	pmdaExtSetFlags(pmda, ap->notify);
+	ap->notify = 0;
     }
     return pmdaFetch(numpmid, pmidlist, resp, pmda);
 }
 
 static int
-mmv_store(pmResult *result, pmdaExt *ep)
+mmv_store(pmResult *result, pmdaExt *pmda)
 {
-    int i, m;
+    agent_t		*ap = (agent_t *)pmdaExtGetData(pmda);
+    int			i, m;
 
-    mmv_reload_maybe(ep);
+    mmv_reload_maybe(pmda);
 
     for (i = 0; i < result->numpmid; i++) {
 	pmValueSet * vsp = result->vset[i];
 
 	if (pmID_cluster(vsp->pmid) == 0) {
-	    for (m = 0; m < mtot; m++) {
+	    for (m = 0; m < ap->mtot; m++) {
 
-		if (pmID_cluster(metrics[m].m_desc.pmid) == 0 && pmID_item(metrics[m].m_desc.pmid) == pmID_item(vsp->pmid)) {
+		if (pmID_cluster(ap->metrics[m].m_desc.pmid) == 0 &&
+		    pmID_item(ap->metrics[m].m_desc.pmid) == pmID_item(vsp->pmid)) {
 		    pmAtomValue atom;
 		    int sts;
 
@@ -1207,7 +1232,7 @@ mmv_store(pmResult *result, pmdaExt *ep)
 					PM_TYPE_32, &atom, PM_TYPE_32)) < 0)
 			return sts;
 		    if (pmID_item(vsp->pmid) == 0)
-			reload = atom.l;
+			ap->reload = atom.l;
 		    else if (pmID_item(vsp->pmid) == 1)
 			pmDebug = atom.l;
 		    else
@@ -1224,34 +1249,40 @@ mmv_store(pmResult *result, pmdaExt *ep)
 static int
 mmv_pmid(const char *name, pmID *pmid, pmdaExt *pmda)
 {
+    agent_t		*ap = (agent_t *)pmdaExtGetData(pmda);
+
     mmv_reload_maybe(pmda);
-    return pmdaTreePMID(pmns, name, pmid);
+    return pmdaTreePMID(ap->pmns, name, pmid);
 }
 
 static int
 mmv_name(pmID pmid, char ***nameset, pmdaExt *pmda)
 {
+    agent_t		*ap = (agent_t *)pmdaExtGetData(pmda);
+
     mmv_reload_maybe(pmda);
-    return pmdaTreeName(pmns, pmid, nameset);
+    return pmdaTreeName(ap->pmns, pmid, nameset);
 }
 
 static int
 mmv_children(const char *name, int traverse, char ***kids, int **sts, pmdaExt *pmda)
 {
+    agent_t		*ap = (agent_t *)pmdaExtGetData(pmda);
+
     mmv_reload_maybe(pmda);
-    return pmdaTreeChildren(pmns, name, traverse, kids, sts);
+    return pmdaTreeChildren(ap->pmns, name, traverse, kids, sts);
 }
 
 static int
-mmv_label_lookup(int ident, int type, pmLabelSet **lp)
+mmv_label_lookup(agent_t *ap, int ident, int type, pmLabelSet **lp)
 {
-    int i, j, id;
-    mmv_disk_label_t lb;
-    stats_t *s;
+    mmv_disk_label_t	lb;
+    stats_t		*s;
+    int			i, j, id;
 
     /* search for labels with requested identifier for given type */
-    for (i = 0; i < scnt; i++) {
-	s = &slist[i];
+    for (i = 0; i < ap->scnt; i++) {
+	s = &ap->slist[i];
 	for (j = 0; j < s->lcnt; j++) {
 	    lb = s->labels[j];
 	    if (type & PM_LABEL_INDOM)
@@ -1268,36 +1299,40 @@ mmv_label_lookup(int ident, int type, pmLabelSet **lp)
 static int
 mmv_label(int ident, int type, pmLabelSet **lp, pmdaExt *pmda)
 {
-    int sts = 0;
+    agent_t		*ap = (agent_t *)pmdaExtGetData(pmda);
+    int			sts = 0;
 
     switch (type) {
 	case PM_LABEL_CLUSTER:
-	    sts = mmv_label_lookup(pmID_cluster(ident), type, lp);
+	    sts = mmv_label_lookup(ap, pmID_cluster(ident), type, lp);
 	    break;
 	case PM_LABEL_INDOM:
-	    sts = mmv_label_lookup(pmInDom_serial(ident), type, lp);
+	    sts = mmv_label_lookup(ap, pmInDom_serial(ident), type, lp);
 	    break;
 	case PM_LABEL_ITEM:
-	    sts = mmv_label_lookup(pmID_item(ident), type, lp);
+	    sts = mmv_label_lookup(ap, pmID_item(ident), type, lp);
 	    break;
 	default:
 	    break;
     } 
     if (sts < 0)
 	return sts;
-    return pmdaLabel(ident, type, lp, pmda);
+    privdata = (void *)ap;
+    sts = pmdaLabel(ident, type, lp, pmda);
+    return sts;
 }
 
 static int
 mmv_labelCallBack(pmInDom indom, unsigned int inst, pmLabelSet **lp)
 {
-    int i, j, count = 0;
-    mmv_disk_label_t lb;
-    stats_t *s;
+    mmv_disk_label_t	lb;
+    stats_t		*s;
+    agent_t		*ap = (agent_t *)privdata;
+    int			i, j, count = 0;
 
     /* search for labels with requested indom and instance identifier */
-    for (i = 0; i < scnt; i++) {
-	s = &slist[i];
+    for (i = 0; i < ap->scnt; i++) {
+	s = &ap->slist[i];
 	for (j = 0; j < s->lcnt; j++) {
 	    lb = s->labels[j];
 	    if (!(lb.flags & PM_LABEL_INSTANCES))
@@ -1314,27 +1349,31 @@ mmv_labelCallBack(pmInDom indom, unsigned int inst, pmLabelSet **lp)
     return count;
 }
 
-void
-__PMDA_INIT_CALL
-mmv_init(pmdaInterface *dp)
+static void
+init_pmda(pmdaInterface *dp, agent_t *ap)
 {
-    int	m;
-    int sep = pmPathSeparator();
+    int		m;
+    int		sep = pmPathSeparator();
+
+    if (!setup) {
+	setup++;
+	fNaN = (float)0.0 / (float)0.0;
+	dNaN = (double)0.0 / (double)0.0;
+	memset(&aNaN, -1, sizeof(pmAtomValue));
+    }
 
     if (isDSO) {
-	pmdaDSO(dp, PMDA_INTERFACE_7, "mmv", NULL);
+	pmdaDSO(dp, PMDA_INTERFACE_7, (char *)ap->prefix, NULL);
     } else {
 	pmSetProcessIdentity(username);
     }
 
-    pcptmpdir = pmGetConfig("PCP_TMP_DIR");
-    pcpvardir = pmGetConfig("PCP_VAR_DIR");
-    pcppmdasdir = pmGetConfig("PCP_PMDAS_DIR");
+    ap->pcptmpdir = pmGetConfig("PCP_TMP_DIR");
+    ap->pcpvardir = pmGetConfig("PCP_VAR_DIR");
+    ap->pcppmdasdir = pmGetConfig("PCP_PMDAS_DIR");
 
-    pmsprintf(statsdir, sizeof(statsdir), "%s%c%s", pcptmpdir, sep, prefix);
-    pmsprintf(pmnsdir, sizeof(pmnsdir), "%s%c" "pmns", pcpvardir, sep);
-    statsdir[sizeof(statsdir)-1] = '\0';
-    pmnsdir[sizeof(pmnsdir)-1] = '\0';
+    pmsprintf(ap->statsdir, MAXPATHLEN, "%s%c%s", ap->pcptmpdir, sep, ap->prefix);
+    pmsprintf(ap->pmnsdir, MAXPATHLEN, "%s%c" "pmns", ap->pcpvardir, sep);
 
     /* Initialize internal dispatch table */
     if (dp->status == 0) {
@@ -1342,23 +1381,18 @@ mmv_init(pmdaInterface *dp)
 	 * number of hard-coded metrics here has to match initializer
 	 * cases below, and pmns initialization in map_stats()
 	 */
-	mtot = 3;
-	if ((metrics = malloc(mtot * sizeof(pmdaMetric))) != NULL) {
+	ap->mtot = 3;
+	if ((ap->metrics = malloc(ap->mtot * sizeof(pmdaMetric))) != NULL) {
 	    /*
 	     * all the hard-coded metrics have the same semantics
 	     */
-	    for (m = 0; m < mtot; m++) {
-		if (m == 0)
-		    metrics[m].m_user = &reload;
-		else if (m == 1)
-		    metrics[m].m_user = &pmDebug;
-		else if (m == 2)
-		    metrics[m].m_user = &scnt;
-		metrics[m].m_desc.pmid = pmID_build(dp->domain, 0, m);
-		metrics[m].m_desc.type = PM_TYPE_32;
-		metrics[m].m_desc.indom = PM_INDOM_NULL;
-		metrics[m].m_desc.sem = PM_SEM_INSTANT;
-		memset(&metrics[m].m_desc.units, 0, sizeof(pmUnits));
+	    for (m = 0; m < ap->mtot; m++) {
+		ap->metrics[m].m_user = (void *)ap;
+		ap->metrics[m].m_desc.pmid = pmID_build(dp->domain, 0, m);
+		ap->metrics[m].m_desc.type = PM_TYPE_32;
+		ap->metrics[m].m_desc.indom = PM_INDOM_NULL;
+		ap->metrics[m].m_desc.sem = PM_SEM_INSTANT;
+		memset(&ap->metrics[m].m_desc.units, 0, sizeof(pmUnits));
 	    }
 	} else {
 	    pmNotifyErr(LOG_ERR, "%s: pmdaInit - out of memory\n",
@@ -1380,16 +1414,48 @@ mmv_init(pmdaInterface *dp)
 	pmdaSetFetchCallBack(dp, mmv_fetchCallBack);
 	pmdaSetLabelCallBack(dp, mmv_labelCallBack);
 
+	pmdaSetData(dp, (void *)ap);
 	pmdaSetFlags(dp, PMDA_EXT_FLAG_HASHED);
-	pmdaInit(dp, indoms, intot, metrics, mtot);
+	pmdaInit(dp, ap->indoms, ap->intot, ap->metrics, ap->mtot);
+    }
+}
+
+void
+__PMDA_INIT_CALL
+pmproxy_init(pmdaInterface *dp)
+{
+    agent_t	*agent = (agent_t *)calloc(1, sizeof(agent_t));
+
+    if (agent) {
+	agent->prefix = pmproxy_prefix;
+	init_pmda(dp, agent);
+    } else {
+	dp->status = -ENOMEM;
+    }
+}
+
+void
+__PMDA_INIT_CALL
+mmv_init(pmdaInterface *dp)
+{
+    agent_t	*agent = (agent_t *)calloc(1, sizeof(agent_t));
+
+    if (agent) {
+	agent->prefix = mmv_prefix;
+	init_pmda(dp, agent);
+    } else {
+	dp->status = -ENOMEM;
     }
 }
 
 int
 main(int argc, char **argv)
 {
+    int		isMMV = 1;
+    int		domain = MMV;
     char	*progname;
     char	logfile[32];
+    const char	*prefix = mmv_prefix;
     pmdaInterface dispatch = { 0 };
 
     isDSO = 0;
@@ -1397,10 +1463,13 @@ main(int argc, char **argv)
     pmGetUsername(&username);
 
     progname = pmGetProgname();
-    if (strncmp(progname, "pmda", 4) == 0 && strlen(progname) > 4)
-	prefix = progname + 4;
+    if (strncmp(progname, "pmda", 4) == 0 && strlen(progname) > 4 &&
+	(isMMV = (strcmp(progname + 4, pmproxy_prefix) != 0)) == 0) {
+	prefix = pmproxy_prefix;
+	domain = PMPROXY;
+    }
     pmsprintf(logfile, sizeof(logfile), "%s.log", prefix);
-    pmdaDaemon(&dispatch, PMDA_INTERFACE_7, progname, MMV, logfile, NULL);
+    pmdaDaemon(&dispatch, PMDA_INTERFACE_7, progname, domain, logfile, NULL);
 
     pmdaGetOptions(argc, argv, &opts, &dispatch);
     if (opts.errors) {
@@ -1411,7 +1480,10 @@ main(int argc, char **argv)
 	username = opts.username;
 
     pmdaOpenLog(&dispatch);
-    mmv_init(&dispatch);
+    if (isMMV)
+	mmv_init(&dispatch);
+    else
+	pmproxy_init(&dispatch);
     pmdaConnect(&dispatch);
     pmdaMain(&dispatch);
     exit(0);
