@@ -232,7 +232,8 @@ static int search_for_config(char *device_path, uint64_t config, char *event_fil
             break;
         }
         if (parsed_config == config) {
-            strncpy(event_file, entry->d_name, strlen(entry->d_name));
+            strncpy(event_file, entry->d_name, MAX_EVENT_NAME-1);
+            event_file[MAX_EVENT_NAME-1] = '\0';	/* buffer overrun guard */
             ret = 0;
             break;
         }
@@ -413,7 +414,7 @@ static int fetch_perf_scale(char *event_name, double *scale)
  */
 static int perf_setup_derived_event(perfdata_t *inst, pmcderived_t *derived_pmc)
 {
-    derived_event_t *curr, *derived_events = inst->derived_events;
+    derived_event_t *curr, *derived_events;
     int nderivedevents = inst->nderivedevents;
     event_t *event;
     pmcsetting_t *derived_setting;
@@ -501,7 +502,7 @@ static int perf_setup_derived_event(perfdata_t *inst, pmcderived_t *derived_pmc)
         return -E_PERFEVENT_LOGIC;
     }
 
-    derived_events = realloc(derived_events,
+    derived_events = realloc(inst->derived_events,
                              (nderivedevents + 1) * sizeof(*derived_events));
     if (NULL == derived_events) {
         free(inst->derived_events);
@@ -523,19 +524,20 @@ static int perf_setup_derived_event(perfdata_t *inst, pmcderived_t *derived_pmc)
 
 /* Setup an event
  */
-static int perf_setup_event(perfdata_t *inst, const char *eventname, const int cpuSetting)
+static int perf_setup_event(perfdata_t *inst, const char *eventname,
+                            unsigned long eventcode, const int cpuSetting)
 {
     int i;
     int ncpus, ret;
     int *cpuarr;
 
-    event_t *events = inst->events;
+    event_t *events;
     int nevents = inst->nevents;
     archinfo_t *archinfo = inst->archinfo;
 
     /* Increase size of event array */
-    events = realloc(events, (nevents + 1) * sizeof(*events) );
-    if(NULL == events)
+    events = realloc(inst->events, (nevents + 1) * sizeof(*events));
+    if (NULL == events)
     {
         free(inst->events);
         inst->nevents = 0;
@@ -607,6 +609,25 @@ static int perf_setup_event(perfdata_t *inst, const char *eventname, const int c
                 continue;
             }
 
+        } else if (!strncmp(eventname, "RAW:", 4)) {
+            info->type = EVENT_TYPE_PERF;
+
+            memset(&info->hw, 0, sizeof(info->hw));
+            info->hw.type = PERF_TYPE_RAW;
+            info->hw.size = sizeof(info->hw);
+            info->hw.config = eventcode;
+            info->hw.read_format = PERF_FORMAT_TOTAL_TIME_ENABLED | PERF_FORMAT_TOTAL_TIME_RUNNING;
+            info->hw.exclude_hv = 1;
+            info->hw.exclude_guest = 1;
+            info->hw.disabled = 1;
+            info->fd = perf_event_open(&info->hw, -1, info->cpu, -1, 0);
+
+            if (info->fd == -1) {
+                fprintf(stderr, "perf_event_open failed on cpu%d for \"%s\": %s\n",
+                        info->cpu, curr->name, strerror(errno));
+                free_eventcpuinfo(info);
+                continue;
+            }
         } else {
 
             info->type = EVENT_TYPE_PERF;
@@ -672,7 +693,7 @@ static int perf_setup_dynamic_events(perfdata_t *inst,
 				     struct pmu *pmu_list)
 {
     int i, ncpus, ret = 0, *cpuarr = NULL, nevents = inst->nevents, *cpumask = NULL;
-    event_t *events = inst->events;
+    event_t *evp, *events = inst->events;
     archinfo_t *archinfo = inst->archinfo;
     struct pmu *pmu_ptr;
     struct pmu_event *event_ptr;
@@ -695,14 +716,15 @@ static int perf_setup_dynamic_events(perfdata_t *inst,
             }
 
             /* Increase the size of event array */
-            events = realloc(events, (nevents + 1) * sizeof(*events));
-            if (!events) {
+            evp = realloc(events, (nevents + 1) * sizeof(*events));
+            if (!evp) {
                 free(inst->events);
                 inst->nevents = 0;
                 inst->events = NULL;
                 free(cpumask);
                 return -E_PERFEVENT_REALLOC;
             }
+	    events = evp;
 
             setup_cpu_config(pmu_ptr, &ncpus, &cpumask);
 
@@ -773,7 +795,6 @@ static int perf_setup_dynamic_events(perfdata_t *inst,
         }
     }
 
-
     inst->events = events;
     inst->nevents = nevents;
     return ret;
@@ -1010,13 +1031,12 @@ static int perf_derived_get(perf_derived_counter **derived_counters,
     perf_derived_counter *pdcounter = *derived_counters;
     int nderivedcounters = *derived_size;
 
-    if(NULL == pdcounter || nderivedcounters != pdata->nderivedevents)
+    if (NULL == pdcounter || nderivedcounters != pdata->nderivedevents)
     {
-        pdcounter = malloc(pdata->nderivedevents * sizeof *pdcounter);
+        pdcounter = calloc(pdata->nderivedevents, sizeof *pdcounter);
         if (NULL == pdcounter) {
             return -E_PERFEVENT_REALLOC;
         }
-        memset(pdcounter, 0, pdata->nderivedevents * sizeof *pdcounter);
         nderivedcounters = pdata->nderivedevents;
 
         for (idx = 0; idx < pdata->nderivedevents; idx++) {
@@ -1032,8 +1052,10 @@ static int perf_derived_get(perf_derived_counter **derived_counters,
                 counter = get_counter(counters, *size, event_list->event->name);
                 if (counter != NULL) {
                     ptr = calloc(1, sizeof(*ptr));
-                    if (!ptr)
+                    if (!ptr) {
+			free(pdcounter);
                         return -E_PERFEVENT_REALLOC;
+		    }
                     ptr->counter = counter;
                     ptr->scale = event_list->scale;
                     ptr->next = NULL;
@@ -1218,14 +1240,15 @@ perfhandle_t *perf_event_create(const char *config_file)
 
     while(pmcsetting)
     {
-        (void) perf_setup_event(inst, pmcsetting->name, pmcsetting->cpuConfig);
+        (void) perf_setup_event(inst, pmcsetting->name, pmcsetting->rawcode,
+                                pmcsetting->cpuConfig);
 
         pmcsetting = pmcsetting->next;
     }
 
     /* Setup the dynamic events */
     if (perfconfig->dynamicpmc) {
-	ret = init_dynamic_events(&pmu_list);
+	ret = init_dynamic_events(&pmu_list, perfconfig->dynamicpmc->dynamicSettingList);
 	if (!ret)
 	    perf_setup_dynamic_events(inst,
 				      perfconfig->dynamicpmc->dynamicSettingList,
