@@ -24,6 +24,7 @@
 char		sep;
 int		vflag;		/* verbose off by default */
 int		nowrap;		/* suppress wrap check */
+int		mflag;		/* check metadata only, suppress pass3 */
 int		index_state = STATE_MISSING;
 int		meta_state = STATE_MISSING;
 int		log_state = STATE_MISSING;
@@ -37,6 +38,7 @@ static pmLongOptions longopts[] = {
     PMAPI_OPTIONS_HEADER("Options"),
     PMOPT_DEBUG,
     { "label", 0, 'l', 0, "print the archive label" },
+    { "metadataonly", 0, 'm', 0, "skip checking log data volumes" },
     PMOPT_NAMESPACE,
     PMOPT_START,
     PMOPT_FINISH,
@@ -50,7 +52,7 @@ static pmLongOptions longopts[] = {
 
 static pmOptions opts = {
     .flags = PM_OPTFLAG_DONE | PM_OPTFLAG_BOUNDARIES | PM_OPTFLAG_STDOUT_TZ,
-    .short_options = "D:ln:S:T:zvwZ:?",
+    .short_options = "D:lmn:S:T:zvwZ:?",
     .long_options = longopts,
     .short_usage = "[options] archive",
 };
@@ -63,6 +65,7 @@ dumpLabel(void)
     char	*yr;
     int		sts;
     char	timebuf[32];	/* for pmCtime result + .xxx */
+    time_t	time;
 
     if ((sts = pmGetArchiveLabel(&label)) < 0) {
 	fprintf(stderr, "%s: cannot get archive label record: %s\n",
@@ -73,7 +76,8 @@ dumpLabel(void)
     fprintf(stderr, "Log Label (Log Format Version %d)\n", label.ll_magic & 0xff);
     fprintf(stderr, "Performance metrics from host %s\n", label.ll_hostname);
 
-    ddmm = pmCtime((const time_t *)&label.ll_start.tv_sec, timebuf);
+    time = label.ll_start.tv_sec;
+    ddmm = pmCtime(&time, timebuf);
     ddmm[10] = '\0';
     yr = &ddmm[20];
     fprintf(stderr, "  commencing %s ", ddmm);
@@ -85,7 +89,8 @@ dumpLabel(void)
         fprintf(stderr, "  ending     UNKNOWN\n");
     }
     else {
-        ddmm = pmCtime((const time_t *)&opts.finish.tv_sec, timebuf);
+	time = opts.finish.tv_sec;
+        ddmm = pmCtime(&time, timebuf);
         ddmm[10] = '\0';
         yr = &ddmm[20];
         fprintf(stderr, "  ending     %s ", ddmm);
@@ -172,6 +177,9 @@ main(int argc, char *argv[])
 	case 'l':	/* display the archive label */
 	    lflag = 1;
 	    break;
+	case 'm':	/* only check metadata */
+	    mflag = 1;
+	    break;
 	case 'v':	/* bump verbosity */
 	    vflag++;
 	    break;
@@ -312,7 +320,8 @@ main(int argc, char *argv[])
 
     sts = pass2(ctxp, archname);
 
-    sts = pass3(ctxp, archname, &opts);
+    if (!mflag)
+	sts = pass3(ctxp, archname, &opts);
 
     if (vflag) {
 	if (result_count > 0)
