@@ -14,6 +14,13 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  * for more details.
+ *
+ * Debug flags
+ * appl0	I/O
+ * appl1	metdata changes
+ * appl2	pmResult changes
+ * appl3	-q and reason for not taking quick exit
+ * appl4	config parser
  */
 
 #include <math.h>
@@ -57,6 +64,7 @@ static pmOptions opts = {
 /*
  *  Global variables
  */
+static int	needti = 0;		/* need time index record */
 static int	first_datarec = 1;	/* first record flag */
 static char	bak_base[MAXPATHLEN+1];	/* basename for backup with -i */
 
@@ -120,6 +128,17 @@ newvolume(int vol)
     else {
 	fprintf(stderr, "%s: __pmLogNewFile(%s,%d) Error: %s\n",
 		pmGetProgname(), outarch.name, vol, pmErrStr(-oserror()));
+	if (oserror() == EEXIST) {
+	    /*
+	     * We've written some files (.index, .meta, .0, ...) and then
+	     * found a duplicate file name ... doesn't matter what you do
+	     * here, badness will result
+	     */
+	    fprintf(stderr, "Removing output files we've created for archive \"%s\"\n", outarch.name);
+	    _pmLogRemove(outarch.name, vol-1);
+	    exit(1);
+	    /*NOTREACHED*/
+	}
 	abandon();
 	/*NOTREACHED*/
     }
@@ -256,6 +275,7 @@ parseargs(int argc, char *argv[])
     int			c;
     int			sts;
     int			sep = pmPathSeparator();
+    char		**cp;
     struct stat		sbuf;
 
     while ((c = pmgetopt_r(argc, argv, &opts)) != EOF) {
@@ -269,9 +289,10 @@ parseargs(int argc, char *argv[])
 		break;
 	    }
 	    if (S_ISREG(sbuf.st_mode) || S_ISLINK(sbuf.st_mode)) {
-		nconf++;
-		if ((conf = (char **)realloc(conf, nconf*sizeof(conf[0]))) != NULL)
-		    conf[nconf-1] = opts.optarg;
+		if ((cp = (char **)realloc(conf, (nconf+1)*sizeof(conf[0]))) != NULL) {
+		    conf = cp;
+		    conf[nconf++] = opts.optarg;
+		}
 	    }
 	    else if (S_ISDIR(sbuf.st_mode)) {
 		DIR		*dirp;
@@ -291,10 +312,10 @@ parseargs(int argc, char *argv[])
 			opts.errors++;
 		    }
 		    else if (S_ISREG(sbuf.st_mode) || S_ISLINK(sbuf.st_mode)) {
-			nconf++;
-			if ((conf = (char **)realloc(conf, nconf*sizeof(conf[0]))) == NULL)
+			if ((cp = (char **)realloc(conf, (nconf+1)*sizeof(conf[0]))) == NULL)
 			    break;
-			if ((conf[nconf-1] = strdup(path)) == NULL) {
+			conf = cp;
+			if ((conf[nconf++] = strdup(path)) == NULL) {
 			    fprintf(stderr, "conf[%d] strdup(%s) failed: %s\n", nconf-1, path, strerror(errno));
 			    abandon();
 			    /*NOTREACHED*/
@@ -644,8 +665,23 @@ reportconfig(void)
     for (lp = label_root; lp != NULL; lp = lp->l_next) {
 	if (lp->flags != 0) {
 	    change |= 1;
-	    printf("\nLabel: %s\n",
+	    printf("\nLabel: %s",
 		   __pmLabelIdentString(lp->old_id, lp->old_type, buf, sizeof(buf)));
+	    if (lp->old_type == PM_LABEL_INSTANCES) {
+		if (lp->old_instance != -1)
+		    printf(", Instance: %d", lp->old_instance);
+		else
+		    printf(", Instances: ALL");
+	    }
+	    if (lp->old_label != NULL)
+		printf(", Label: %s", lp->old_label);
+	    if (lp->old_value != NULL)
+		printf(", Value: %s", lp->old_value);
+	    putchar ('\n');
+	}
+	if (lp->flags & LABEL_NEW) {
+	    printf("NEW:\n");
+	    pmPrintLabelSets(stdout, lp->new_id, lp->new_type, lp->new_labels, 1);
 	}
 	if (lp->flags & LABEL_CHANGE_ID) {
 	    printf("ID:\t\t%s -> ", 
@@ -653,10 +689,16 @@ reportconfig(void)
 	    printf("%s\n", 
 		   labelIDStr(lp->new_type, lp->new_id, buf, sizeof(buf)));
 	}
-	if (lp->flags & LABEL_CHANGE_LABEL)
-	    printf("Label:\t\t\"%s\" -> \"%s\n", lp->old_label, lp->new_label); 
-	if (lp->flags & LABEL_CHANGE_VALUE)
-	    printf("Label:\t\t\"value\" -> \"%s\n", lp->new_value); 
+	if (lp->flags & LABEL_CHANGE_LABEL) {
+	    printf("Label:\t\t%s -> %s\n",
+		   lp->old_label ? lp->old_label : "ALL",
+		   lp->new_label);
+	}
+	if (lp->flags & LABEL_CHANGE_VALUE) {
+	    printf("Value:\t\t%s -> %s\n",
+		   lp->old_value ? lp->old_value : "ALL",
+		   lp->new_value);
+	}
 	if (lp->flags & LABEL_DELETE)
 	    printf("DELETE\n");
     }
@@ -673,27 +715,114 @@ anychange(void)
     const labelspec_t	*lp;
     int			i;
 
-    if (global.flags != 0)
+    if (global.flags != 0) {
+	if (pmDebugOptions.appl3) {
+	    fprintf(stderr, "anychange: global.flags (%d,", global.flags);
+	    if (global.flags & GLOBAL_CHANGE_TIME)
+		fprintf(stderr, " CHANGE_TIME");
+	    if (global.flags & GLOBAL_CHANGE_HOSTNAME)
+		fprintf(stderr, " CHANGE_HOSTNAME");
+	    if (global.flags & GLOBAL_CHANGE_TZ)
+		fprintf(stderr, " CHANGE_TZ");
+	    fprintf(stderr, ") != 0\n");
+	}
 	return 1;
+    }
     for (ip = indom_root; ip != NULL; ip = ip->i_next) {
-	if (ip->new_indom != ip->old_indom)
+	if (ip->new_indom != ip->old_indom) {
+	    if (pmDebugOptions.appl3) {
+		fprintf(stderr, "anychange: indom %s changed\n", pmInDomStr(ip->old_indom));
+	    }
 	    return 1;
+	}
 	for (i = 0; i < ip->numinst; i++) {
-	    if (ip->inst_flags[i])
+	    if (ip->inst_flags[i]) {
+		if (pmDebugOptions.appl3) {
+		    fprintf(stderr, "anychange: indom %s inst %d flags (%d,",
+			pmInDomStr(ip->old_indom), i, ip->inst_flags[i]);
+		    if (ip->inst_flags[i] & INST_CHANGE_INST)
+			fprintf(stderr, " CHANGE_INST");
+		    if (ip->inst_flags[i] & INST_CHANGE_INAME)
+			fprintf(stderr, " CHANGE_INAME");
+		    if (ip->inst_flags[i] & INST_DELETE)
+			fprintf(stderr, " DELETE");
+		    fprintf(stderr, ") != 0\n");
+		}
 		return 1;
+	    }
 	}
     }
     for (mp = metric_root; mp != NULL; mp = mp->m_next) {
-	if (mp->flags != 0 || mp->ip != NULL)
+	if (mp->flags != 0 || mp->ip != NULL) {
+	    if (pmDebugOptions.appl3) {
+		if (mp->flags != 0) {
+		    fprintf(stderr, "anychange: metric %s flags (%d,",
+			mp->old_name, mp->flags);
+		    if (mp->flags & METRIC_CHANGE_PMID)
+			fprintf(stderr, " CHANGE_PMID");
+		    if (mp->flags & METRIC_CHANGE_NAME)
+			fprintf(stderr, " CHANGE_NAME");
+		    if (mp->flags & METRIC_CHANGE_TYPE)
+			fprintf(stderr, " CHANGE_TYPE");
+		    if (mp->flags & METRIC_CHANGE_INDOM)
+			fprintf(stderr, " CHANGE_INDOM");
+		    if (mp->flags & METRIC_CHANGE_SEM)
+			fprintf(stderr, " CHANGE_SEM");
+		    if (mp->flags & METRIC_CHANGE_UNITS)
+			fprintf(stderr, " CHANGE_UNITS");
+		    if (mp->flags & METRIC_DELETE)
+			fprintf(stderr, " DELETE");
+		    if (mp->flags & METRIC_RESCALE)
+			fprintf(stderr, " RESCALE");
+		    fprintf(stderr, ") != 0\n");
+		}
+		if (mp->ip != NULL)
+		    fprintf(stderr, "anychange: metric %s ip NULL\n", mp->old_name);
+	    }
 	    return 1;
+	}
     }
     for (lp = label_root; lp != NULL; lp = lp->l_next) {
-	if (lp->flags != 0 || lp->ip != NULL)
+	if (lp->flags != 0 || lp->ip != NULL) {
+	    if (pmDebugOptions.appl3) {
+		if (lp->flags != 0) {
+		    fprintf(stderr, "anychange: label %s flags (%d,",
+			lp->old_label, lp->flags);
+		    if (lp->flags & LABEL_ACTIVE)
+			fprintf(stderr, " ACTIVE");
+		    if (lp->flags & LABEL_CHANGE_ID)
+			fprintf(stderr, " CHANGE_ID");
+		    if (lp->flags & LABEL_CHANGE_LABEL)
+			fprintf(stderr, " CHANGE_LABEL");
+		    if (lp->flags & LABEL_CHANGE_INSTANCE)
+			fprintf(stderr, " CHANGE_INSTANCE");
+		    if (lp->flags & LABEL_DELETE)
+			fprintf(stderr, " DELETE");
+		    if (lp->flags & LABEL_NEW)
+			fprintf(stderr, " NEW");
+		    fprintf(stderr, ") != 0\n");
+		}
+		if (lp->ip != NULL)
+		    fprintf(stderr, "anychange: label %s ip NULL\n", lp->old_label);
+	    }
 	    return 1;
+	}
     }
     for (tp = text_root; tp != NULL; tp = tp->t_next) {
-	if (tp->flags != 0 || tp->ip != NULL)
+	if (tp->flags != 0 || tp->ip != NULL) {
+	    if (pmDebugOptions.appl3) {
+		if (tp->flags != 0)
+		    fprintf(stderr, "anychange: %s %s text flags (%d) != 0\n",
+		       (tp->old_type & PM_TEXT_PMID) ? "pmID" : "pmInDom",
+		       (tp->old_type & PM_TEXT_PMID) ? pmIDStr(tp->old_id) : pmInDomStr(tp->old_id),
+		       tp->flags);
+		if (tp->ip != NULL)
+		    fprintf(stderr, "anychange: %s %s text ip NULL\n",
+		       (tp->old_type & PM_TEXT_PMID) ? "pmID" : "pmInDom",
+		       (tp->old_type & PM_TEXT_PMID) ? pmIDStr(tp->old_id) : pmInDomStr(tp->old_id));
+	    }
 	    return 1;
+	}
     }
     
     return 0;
@@ -735,17 +864,19 @@ fixstamp(struct timeval *tvp)
 static void
 link_entries(void)
 {
-    indomspec_t		*ip;
-    metricspec_t	*mp;
-    textspec_t		*tp;
-    labelspec_t		*lp;
-    __pmHashCtl		*hcp, *hcp2;
-    __pmHashNode	*node, *node2;
-    int			old_id, new_id;
-    int			i;
-    int			type;
-    int			change;
-    char		strbuf[64];
+    indomspec_t			*ip;
+    metricspec_t		*mp;
+    textspec_t			*tp;
+    labelspec_t			*lp;
+    const pmLabelSet		*lsp;
+    const __pmLogLabelSet	*llsp;
+    __pmHashCtl			*hcp, *hcp2;
+    __pmHashNode		*node, *node2;
+    int				old_id, new_id;
+    int				i;
+    int				type;
+    int				change;
+    char			strbuf[64];
 
     /* Link metricspec_t entries to indomspec_t entries */
     hcp = &inarch.ctxp->c_archctl->ac_log->l_hashpmid;
@@ -807,7 +938,7 @@ link_entries(void)
 		    continue;
 
 		/* Found one. */
-		tp = start_text(type, (int)(node2->key));
+		tp = start_text(type, (int)(node2->key), NULL);
 		assert(tp->old_id == ip->old_indom);
 		if (change)
 		    tp->ip = ip;
@@ -852,7 +983,7 @@ link_entries(void)
 		    continue;
 
 		/* Found one. */
-		tp = start_text(type, (int)(node2->key));
+		tp = start_text(type, (int)(node2->key), NULL);
 		assert(tp->old_id == mp->old_desc.pmid);
 		if (mp->new_desc.pmid != mp->old_desc.pmid) {
 		    if (tp->flags & TEXT_CHANGE_ID) {
@@ -898,22 +1029,26 @@ link_entries(void)
 		    continue;
 
 		/* Found one. */
-		lp = start_label(type, (int)(node2->key), NULL);
-		assert(lp->old_id == ip->old_indom);
-		if (change)
-		    lp->ip = ip;
-		if (ip->new_indom != ip->old_indom) {
-		    if (lp->flags & LABEL_CHANGE_ID) {
-			/* indom already changed via label clause */
-			if (lp->new_id != ip->new_indom) {
-			    pmsprintf(strbuf, sizeof(strbuf), "%s", pmInDomStr(lp->new_id));
-			    pmsprintf(mess, sizeof(mess), "Conflicting indom change for label set (%s from text clause, %s from indom clause)", strbuf, pmInDomStr(ip->new_indom));
-			    yysemantic(mess);
+		llsp = (__pmLogLabelSet *)node2->data;
+		for (i = 0; i < llsp->nsets; ++i) {
+		    lsp = &llsp->labelsets[i];
+		    lp = start_label(type, (int)node2->key, lsp->inst, NULL, NULL, NULL);
+		    assert(lp->old_id == ip->old_indom);
+		    if (change)
+			lp->ip = ip;
+		    if (ip->new_indom != ip->old_indom) {
+			if (lp->flags & LABEL_CHANGE_ID) {
+			    /* indom already changed via label clause */
+			    if (lp->new_id != ip->new_indom) {
+				pmsprintf(strbuf, sizeof(strbuf), "%s", pmInDomStr(lp->new_id));
+				pmsprintf(mess, sizeof(mess), "Conflicting indom change for label set (%s from text clause, %s from indom clause)", strbuf, pmInDomStr(ip->new_indom));
+				yysemantic(mess);
+			    }
+			}
+			else {
+			    lp->flags |= LABEL_CHANGE_ID;
+			    lp->new_id = ip->new_indom;
 			}
-		    }
-		    else {
-			lp->flags |= LABEL_CHANGE_ID;
-			lp->new_id = ip->new_indom;
 		    }
 		}
 	    }
@@ -957,7 +1092,7 @@ link_entries(void)
 		    continue;
 
 		/* Found one. */
-		lp = start_label(type, old_id, NULL);
+		lp = start_label(type, old_id, 0, NULL, NULL, NULL);
 		assert(lp->old_id == old_id);
 		if (old_id != new_id) {
 		    if (lp->flags & LABEL_CHANGE_ID) {
@@ -1253,6 +1388,90 @@ check_output()
     }
 }
 
+static void
+do_newlabelsets(void)
+{
+    long		out_offset;
+    unsigned int	type;
+    unsigned int	ident;
+    int			nsets;
+    pmLabelSet		*labellist = NULL;
+    pmTimeval		stamp;
+    labelspec_t		*lp;
+    int			sts;
+    char		buf[64];
+
+    out_offset = __pmFtell(outarch.logctl.l_mdfp);
+
+    /*
+     * Traverse the list of label change records and emit any new label sets
+     * at the globally adjusted start time.
+     */
+    stamp.tv_sec = inarch.rp->timestamp.tv_sec;
+    stamp.tv_usec = inarch.rp->timestamp.tv_usec;
+    
+    for (lp = label_root; lp != NULL; lp = lp->l_next) {
+	/* Is this a new label record? */
+	if (! ((lp->flags & LABEL_NEW)))
+	    continue;
+
+	/*
+	 * Write the record.
+	 * libpcp, via __pmLogPutLabel(), assumes control of the storage pointed
+	 * to by labellist.
+	 */
+	ident = lp->new_id;
+	type = lp->new_type;
+	nsets = 1;
+	labellist = lp->new_labels;
+
+	if (pmDebugOptions.appl1) {
+	    fprintf(stderr, "New: labels for ");
+	    if ((lp->old_type & PM_LABEL_CONTEXT))
+		fprintf(stderr, " context\n");
+	    else if ((lp->old_type & PM_LABEL_DOMAIN))
+		fprintf(stderr, " domain %d\n", pmID_domain(lp->old_id));
+	    else if ((lp->old_type & PM_LABEL_CLUSTER))
+		fprintf(stderr, " cluster %d.%d\n", pmID_domain(lp->old_id), pmID_cluster (lp->old_id));
+	    else if ((lp->old_type & PM_LABEL_ITEM))
+		fprintf(stderr, " item %s\n", pmIDStr(lp->old_id));
+	    else if ((lp->old_type & PM_LABEL_INDOM))
+		fprintf(stderr, " indom %s\n", pmInDomStr(lp->old_id));
+	    else if ((lp->old_type & PM_LABEL_INSTANCES))
+		fprintf(stderr, " the instances of indom %s\n", pmInDomStr(lp->old_id));
+	    pmPrintLabelSets(stderr, ident, type, labellist, nsets);
+	}
+
+	if ((sts = __pmLogPutLabel(&outarch.archctl, type, ident,
+				   nsets, labellist, &stamp)) < 0) {
+	    fprintf(stderr, "%s: Error: __pmLogPutLabel: %s: %s\n",
+		    pmGetProgname(),
+		    __pmLabelIdentString(ident, type, buf, sizeof(buf)),
+		    pmErrStr(sts));
+	    abandon();
+	    /*NOTREACHED*/
+	}
+
+	if (pmDebugOptions.appl0) {
+	    fprintf(stderr, "Metadata: write LabelSet %s @ offset=%ld\n",
+		    __pmLabelIdentString(ident, type, buf, sizeof(buf)), out_offset);
+	}
+
+	if (first_datarec) {
+	    first_datarec = 0;
+	    /*
+	     * Any global time adjustment done after the first record is output
+	     * above
+	     */
+	    outarch.logctl.l_label.ill_start.tv_sec = inarch.rp->timestamp.tv_sec;
+	    outarch.logctl.l_label.ill_start.tv_usec = inarch.rp->timestamp.tv_usec;
+	    /* need to fix start-time in label records */
+	    writelabel(1);
+	    needti = 1;
+	}
+    }
+}
+
 int
 main(int argc, char **argv)
 {
@@ -1262,7 +1481,6 @@ main(int argc, char **argv)
     int		i;
     int		ti_idx;			/* next slot for input temporal index */
     int		dir_fd = -1;		/* poinless initialization to humour gcc */
-    int		needti = 0;
     int		doneti = 0;
     pmTimeval	tstamp = { 0 };		/* for last log record */
     off_t	old_log_offset = 0;	/* log offset before last log record */
@@ -1442,15 +1660,23 @@ main(int argc, char **argv)
     if (Cflag)
 	exit(0);
 
-    if (qflag && anychange() == 0)
+    if (qflag && anychange() == 0) {
+	if (pmDebugOptions.appl3) {
+	    fprintf(stderr, "Done, no rewriting required\n");
+	}
 	exit(0);
+    }
 
     /* create output log - must be done before writing label */
     outarch.archctl.ac_log = &outarch.logctl;
     if ((sts = __pmLogCreate("", outarch.name, PM_LOG_VERS02, &outarch.archctl)) < 0) {
 	fprintf(stderr, "%s: Error: __pmLogCreate(%s): %s\n",
 		pmGetProgname(), outarch.name, pmErrStr(sts));
-	abandon();
+	/*
+	 * do not cleanup ... if error is EEXIST we should not clobber an
+	 * existing (and persumably) good archive
+	 */
+	exit(1);
 	/*NOTREACHED*/
     }
 
@@ -1525,6 +1751,13 @@ main(int argc, char **argv)
 	 * */
 	fixstamp(&inarch.rp->timestamp);
 
+	/*
+	 * Write out any new label sets before any other data using the adjusted
+	 * time stamp of the first data record.
+	 */
+	if (first_datarec)
+	    do_newlabelsets();
+
 	/*
 	 * process metadata until find an indom record with timestamp
 	 * after the current log record, or a metric record for a pmid
@@ -1729,7 +1962,7 @@ main(int argc, char **argv)
 	    abandon();
 	    /*NOTREACHED*/
 	}
-	_pmLogRemove(bak_base);
+	_pmLogRemove(bak_base, -1);
     }
 
     exit(0);
@@ -1739,11 +1972,12 @@ void
 abandon(void)
 {
     char    path[MAXNAMELEN+1];
+
     if (dflag == 0) {
 	if (Cflag == 0 && iflag == 0)
 	    fprintf(stderr, "Archive \"%s\" not created.\n", outarch.name);
 
-	_pmLogRemove(outarch.name);
+	_pmLogRemove(outarch.name, -1);
 	if (iflag)
 	    _pmLogRename(bak_base, inarch.name);
 	while (outarch.archctl.ac_curvol >= 0) {
@@ -1760,4 +1994,5 @@ abandon(void)
 	fprintf(stderr, "Archive \"%s\" creation truncated.\n", outarch.name);
 
     exit(1);
+    /*NOTREACHED*/
 }
