@@ -1,6 +1,6 @@
 #!/usr/bin/env pmpython
 #
-# Copyright (C) 2015-2018 Marko Myllynen <myllynen@redhat.com>
+# Copyright (C) 2015-2019 Marko Myllynen <myllynen@redhat.com>
 #
 # This program is free software; you can redistribute it and/or modify it
 # under the terms of the GNU General Public License as published by the
@@ -29,7 +29,6 @@ import sys
 
 # Our imports
 from datetime import datetime, timedelta
-import socket
 import time
 import math
 import re
@@ -38,9 +37,8 @@ import os
 # PCP Python PMAPI
 from pcp import pmapi, pmi, pmconfig
 from cpmapi import PM_CONTEXT_ARCHIVE, PM_CONTEXT_LOCAL
-from cpmapi import PM_ERR_EOL, PM_IN_NULL, PM_DEBUG_APPL1
+from cpmapi import PM_IN_NULL, PM_DEBUG_APPL1, PM_TIME_SEC
 from cpmapi import PM_SEM_DISCRETE, PM_TYPE_FLOAT, PM_TYPE_DOUBLE, PM_TYPE_STRING
-from cpmapi import PM_TIME_SEC
 from cpmi import PMI_ERR_DUPINSTNAME
 
 if sys.version_info[0] >= 3:
@@ -76,15 +74,16 @@ class PMReporter(object):
         # Configuration directives
         self.keys = ('source', 'output', 'derived', 'header', 'globals',
                      'samples', 'interval', 'type', 'precision', 'daemonize',
-                     'timestamp', 'unitinfo', 'colxrow', 'separate_header',
+                     'timestamp', 'unitinfo', 'colxrow', 'separate_header', 'fixed_header',
                      'delay', 'width', 'delimiter', 'extcsv', 'width_force',
                      'extheader', 'repeat_header', 'timefmt', 'interpol',
-                     'dynamic_header', 'overall_rank', 'overall_rank_alt',
+                     'dynamic_header', 'overall_rank', 'overall_rank_alt', 'sort_metric',
                      'count_scale', 'space_scale', 'time_scale', 'version',
                      'count_scale_force', 'space_scale_force', 'time_scale_force',
                      'type_prefer', 'precision_force', 'limit_filter', 'limit_filter_force',
-                     'live_filter', 'rank', 'invert_filter', 'predicate',
-                     'speclocal', 'instances', 'ignore_incompat', 'omit_flat')
+                     'live_filter', 'rank', 'invert_filter', 'predicate', 'names_change',
+                     'speclocal', 'instances', 'ignore_incompat', 'ignore_unknown',
+                     'omit_flat')
 
         # The order of preference for options (as present):
         # 1 - command line options
@@ -107,6 +106,8 @@ class PMReporter(object):
         self.type = 0
         self.type_prefer = self.type
         self.ignore_incompat = 0
+        self.ignore_unknown = 0
+        self.names_change = 0 # ignore
         self.instances = []
         self.live_filter = 0
         self.rank = 0
@@ -116,6 +117,7 @@ class PMReporter(object):
         self.limit_filter_force = 0
         self.invert_filter = 0
         self.predicate = None
+        self.sort_metric = None
         self.omit_flat = 0
         self.colxrow = None
         self.width = 0
@@ -125,6 +127,7 @@ class PMReporter(object):
         self.delimiter = None
         self.extcsv = 0
         self.extheader = 0
+        self.fixed_header = 0
         self.repeat_header = 0
         self.dynamic_header = 0
         self.separate_header = 0
@@ -150,6 +153,7 @@ class PMReporter(object):
         self.runtime = -1
         self.found_insts = []
         self.prev_insts = None
+        self.static_header = 1
 
         # Performance metrics store
         # key - metric name
@@ -171,7 +175,7 @@ class PMReporter(object):
         opts = pmapi.pmOptions()
         opts.pmSetOptionCallback(self.option)
         opts.pmSetOverrideCallback(self.option_override)
-        opts.pmSetShortOptions("a:h:LK:c:Co:F:e:D:V?HUGpA:S:T:O:s:t:Z:zdrRIi:jJ:238:9:nN:vX:W:w:P:0:l:kxE:1gf:uq:b:y:Q:B:Y:")
+        opts.pmSetShortOptions("a:h:LK:c:Co:F:e:D:V?HUGpA:S:T:O:s:t:Z:zdrRIi:jJ:234:578:9:nN:6:vX:W:w:P:0:l:kxE:1gf:uq:b:y:Q:B:Y:")
         opts.pmSetShortUsage("[option...] metricspec [...]")
 
         opts.pmSetLongOptionHeader("General options")
@@ -186,7 +190,7 @@ class PMReporter(object):
         opts.pmSetLongOption("output", 1, "o", "OUTPUT", "output target: archive, csv, stdout (default)")
         opts.pmSetLongOption("output-file", 1, "F", "OUTFILE", "output file")
         opts.pmSetLongOption("derived", 1, "e", "FILE|DFNT", "derived metrics definitions")
-        self.daemonize = opts.pmSetLongOption("daemonize", 0, "", "", "daemonize on startup") # > 1
+        opts.pmSetLongOption("daemonize", 0, "", "", "daemonize on startup")
         opts.pmSetLongOptionDebug()        # -D/--debug
         opts.pmSetLongOptionVersion()      # -V/--version
         opts.pmSetLongOptionHelp()         # -?/--help
@@ -208,15 +212,18 @@ class PMReporter(object):
         opts.pmSetLongOption("raw", 0, "r", "", "output raw counter values (no rate conversion)")
         opts.pmSetLongOption("raw-prefer", 0, "R", "", "prefer output raw counter values (no rate conversion)")
         opts.pmSetLongOption("ignore-incompat", 0, "I", "", "ignore incompatible instances (default: abort)")
+        opts.pmSetLongOption("ignore-unknown", 0, "5", "", "ignore unknown metrics (default: abort)")
         opts.pmSetLongOption("instances", 1, "i", "STR", "instances to report (default: all current)")
         opts.pmSetLongOption("live-filter", 0, "j", "", "perform instance live filtering")
         opts.pmSetLongOption("rank", 1, "J", "COUNT", "limit results to COUNT highest/lowest valued instances")
         opts.pmSetLongOption("overall-rank", 0, "2", "", "report overall ranking from archive")
         opts.pmSetLongOption("overall-rank-alt", 0, "3", "", "report overall ranking from archive in pmrep format")
+        opts.pmSetLongOption("names-change", 1, "4", "ACTION", "update/ignore/abort on PMNS changes (default: ignore)")
         opts.pmSetLongOption("limit-filter", 1, "8", "LIMIT", "default limit for value filtering")
         opts.pmSetLongOption("limit-filter-force", 1, "9", "LIMIT", "forced limit for value filtering")
         opts.pmSetLongOption("invert-filter", 0, "n", "", "perform ranking before live filtering")
         opts.pmSetLongOption("predicate", 1, "N", "METRIC", "set predicate filter reference metric")
+        opts.pmSetLongOption("sort-metric", 1, "6", "METRIC", "set sort reference metric for colxrow output")
         opts.pmSetLongOption("omit-flat", 0, "v", "", "omit single-valued metrics")
         opts.pmSetLongOption("colxrow", 1, "X", "STR", "swap stdout columns and rows using STR as header label")
         opts.pmSetLongOption("width", 1, "w", "N", "default column width")
@@ -226,8 +233,9 @@ class PMReporter(object):
         opts.pmSetLongOption("delimiter", 1, "l", "STR", "delimiter to separate csv/stdout columns")
         opts.pmSetLongOption("extended-csv", 0, "k", "", "write extended CSV")
         opts.pmSetLongOption("extended-header", 0, "x", "", "display extended header")
+        opts.pmSetLongOption("fixed-header", 0, "7", "", "fixed header and dynamic instances with colxrow output")
         opts.pmSetLongOption("repeat-header", 1, "E", "N", "repeat stdout headers every N lines")
-        opts.pmSetLongOption("dynamic-header", 0, "1", "", "update header dynamically when instances dis/appear")
+        opts.pmSetLongOption("dynamic-header", 0, "1", "", "update header dynamically on metric/instance changes")
         opts.pmSetLongOption("separate-header", 0, "g", "", "write separated header before metrics")
         opts.pmSetLongOption("timestamp-format", 1, "f", "STR", "strftime string for timestamp format")
         opts.pmSetLongOption("no-interpol", 0, "u", "", "disable interpolation mode with archives")
@@ -246,12 +254,11 @@ class PMReporter(object):
             return 1
         return 0
 
-    def option(self, opt, optarg, index):
-        """ Perform setup for an individual command line option """
-        if index == self.daemonize and opt == '':
+    def option(self, opt, optarg, _index):
+        """ Perform setup for individual command line option """
+        if opt == 'daemonize':
             self.daemonize = 1
-            return
-        if opt == 'K':
+        elif opt == 'K':
             if not self.speclocal or not self.speclocal.startswith(";"):
                 self.speclocal = ";" + optarg
             else:
@@ -292,6 +299,8 @@ class PMReporter(object):
             self.type_prefer = 1
         elif opt == 'I':
             self.ignore_incompat = 1
+        elif opt == '5':
+            self.ignore_unknown = 1
         elif opt == 'i':
             self.instances = self.instances + self.pmconfig.parse_instances(optarg)
         elif opt == 'j':
@@ -302,6 +311,16 @@ class PMReporter(object):
             self.overall_rank = 1
         elif opt == '3':
             self.overall_rank_alt = 1
+        elif opt == '4':
+            if optarg == 'ignore':
+                self.names_change = 0
+            elif optarg == 'abort':
+                self.names_change = 1
+            elif optarg == 'update':
+                self.names_change = 2
+            else:
+                sys.stderr.write("Unknown names-change action '%s' specified.\n" % optarg)
+                sys.exit(1)
         elif opt == '8':
             self.limit_filter = optarg
         elif opt == '9':
@@ -310,6 +329,8 @@ class PMReporter(object):
             self.invert_filter = 1
         elif opt == 'N':
             self.predicate = optarg
+        elif opt == '6':
+            self.sort_metric = optarg
         elif opt == 'v':
             self.omit_flat = 1
         elif opt == 'X':
@@ -330,6 +351,8 @@ class PMReporter(object):
             self.extheader = 1
         elif opt == 'E':
             self.repeat_header = optarg
+        elif opt == '7':
+            self.fixed_header = 1
         elif opt == '1':
             self.dynamic_header = 1
         elif opt == 'g':
@@ -354,7 +377,7 @@ class PMReporter(object):
             raise pmapi.pmUsageErr()
 
     def connect(self):
-        """ Establish a PMAPI context """
+        """ Establish PMAPI context """
         context, self.source = pmapi.pmContext.set_connect_options(self.opts, self.source, self.speclocal)
 
         self.pmfg = pmapi.fetchgroup(context, self.source)
@@ -392,11 +415,11 @@ class PMReporter(object):
                 sys.stderr.write("Output directory %s not accessible.\n" % outdir)
                 sys.exit(1)
 
-        # Metric text label width can be ignored
+        # Set default width when needed
         if self.separate_header and not self.width:
             self.width = 8
 
-        # Adjustments and checks for for overall rankings
+        # Adjustments and checks for overall rankings
         if not self.rank and (self.overall_rank or self.overall_rank_alt):
             sys.stderr.write("Overall ranking requires ranking enabled.\n")
             sys.exit(1)
@@ -411,9 +434,36 @@ class PMReporter(object):
             self.colxrow = None
             self.predicate = None
 
+        # Adjust header selection
+        if self.colxrow is None or self.output != OUTPUT_STDOUT:
+            self.fixed_header = 0
+        if self.dynamic_header:
+            self.fixed_header = 0
+        if self.fixed_header:
+            self.dynamic_header = 0
+
+        if self.names_change == 2:
+            self.fixed_header = 0
+            self.dynamic_header = 1
+
+        self.static_header = 1 not in (self.fixed_header, self.dynamic_header)
+
         self.pmconfig.validate_metrics(curr_insts=not self.live_filter)
         self.pmconfig.finalize_options()
 
+        if self.sort_metric:
+            for sort_metric in self.sort_metric.split(","):
+                sort_metric = sort_metric[1:] if sort_metric[:1] == "-" else sort_metric
+                if sort_metric not in self.metrics:
+                    sys.stderr.write("Sort reference metric %s not part of metrics.\n" % sort_metric)
+                    sys.exit(1)
+
+                i = list(self.metrics.keys()).index(sort_metric)
+
+                if self.pmconfig.insts[i][0][0] == pmapi.c_api.PM_IN_NULL:
+                    sys.stderr.write("Sort reference metric must have instances.\n")
+                    sys.exit(1)
+
     def execute(self):
         """ Fetch and report """
         # Debug
@@ -474,8 +524,14 @@ class PMReporter(object):
             time.sleep(align)
 
         # Main loop
+        refresh_metrics = 0
         while self.samples != 0:
-            # Repeat static header if needed
+            # Refresh metrics as needed
+            if refresh_metrics:
+                refresh_metrics = 0
+                self.pmconfig.update_metrics(curr_insts=not self.live_filter)
+
+            # Repeat header if needed
             if self.output == OUTPUT_STDOUT and not self.dynamic_header:
                 if self.lines > 0 and self.repeat_header == self.lines:
                     self.write_header(True)
@@ -483,17 +539,9 @@ class PMReporter(object):
                 self.lines += 1
 
             # Fetch values
-            try:
-                self.pmfg.fetch()
-            except pmapi.pmErr as error:
-                if error.args[0] == PM_ERR_EOL:
-                    break
-                raise error
-
-            # Watch for endtime in uninterpolated mode
-            if not self.interpol:
-                if float(self.pmfg_ts().strftime('%s')) > float(self.opts.pmGetOptionFinish()):
-                    break
+            refresh_metrics = self.pmconfig.fetch()
+            if refresh_metrics < 0:
+                break
 
             # Report and prepare for the next round
             self.report(self.pmfg_ts())
@@ -506,8 +554,8 @@ class PMReporter(object):
         self.report(None)
 
     def report(self, tstamp):
-        """ Report the metric values """
-        if tstamp != None:
+        """ Report metric values """
+        if tstamp is not None:
             tstamp = tstamp.strftime(self.timefmt)
 
         if self.overall_rank:
@@ -592,7 +640,7 @@ class PMReporter(object):
         # Metrics / text labels
         self.labels = OrderedDict() # pylint: disable=attribute-defined-outside-init
         for i, metric in enumerate(self.metrics):
-            if results and not results[metric]:
+            if self.dynamic_header and results and not results[metric]:
                 continue
             l = str(self.metrics[metric][4])
             label = self.metrics[metric][0]
@@ -613,7 +661,7 @@ class PMReporter(object):
         self.format = self.format[:-l]
 
         # Collect the instances in play
-        if not self.dynamic_header:
+        if self.static_header:
             for i in range(len(self.metrics)):
                 for instance in self.pmconfig.insts[i][1]:
                     if instance not in self.found_insts:
@@ -653,7 +701,7 @@ class PMReporter(object):
 
         instances = sum([len(x[0]) for x in self.pmconfig.insts])
         insts_txt = "instances" if instances != 1 else "instance"
-        if self.dynamic_header:
+        if not self.static_header:
             if self.context.type == PM_CONTEXT_ARCHIVE:
                 insts_txt += " present in archive"
             else:
@@ -666,7 +714,7 @@ class PMReporter(object):
         def secs_to_readable(seconds):
             """ Convert seconds to easily readable format """
             seconds = float(math.floor((seconds) + math.copysign(0.5, seconds)))
-            parts = str(timedelta(seconds=int(round(seconds)))).split(':')
+            parts = str(timedelta(seconds=int(round(seconds)))).split(":")
             if len(parts[0]) == 1:
                 parts[0] = "0" + parts[0]
             elif parts[0][-2] == " ":
@@ -678,7 +726,7 @@ class PMReporter(object):
             if not self.interpol and self.opts.pmGetOptionSamples():
                 samples = str(samples) + " (requested)"
             elif not self.interpol:
-                samples = "N/A" # pylint: disable=redefined-variable-type
+                samples = "N/A"
 
         comm = "#" if self.output == OUTPUT_CSV else ""
         self.writer.write(comm + "\n")
@@ -704,8 +752,8 @@ class PMReporter(object):
         return zip(range(l), r)
 
     def get_instance_count(self, results):
-        """ Helper to get the number of instances of current results """
-        if not self.dynamic_header:
+        """ Helper to get number of instances of current results """
+        if self.static_header:
             if self.colxrow is None:
                 c = len(str(sum([len(i[0]) for i in self.pmconfig.insts])))
             else:
@@ -726,10 +774,12 @@ class PMReporter(object):
             line += metric
             if self.pmconfig.insts[i][0][0] != PM_IN_NULL and name:
                 line += "[\"" + name + "\"]"
-            if self.metrics[metric][2][0]:
-                line += " - " + self.metrics[metric][2][0] + "\n"
-            else:
-                line += " - none\n"
+            if self.unitinfo:
+                if self.metrics[metric][2][0]:
+                    line += " - " + self.metrics[metric][2][0]
+                else:
+                    line += " - none"
+            line += "\n"
             self.writer.write(line.format(str(k)))
 
         k = 0
@@ -737,15 +787,18 @@ class PMReporter(object):
             for i, metric in enumerate(self.metrics):
                 for j, n in self.get_results_iter(i, metric, results):
                     k += 1
-                    name = self.pmconfig.insts[i][1][j] if not self.dynamic_header else str(n[1])
+                    name = self.pmconfig.insts[i][1][j] if self.static_header else str(n[1])
                     write_line(metric, k, i, name)
         else:
             for label in self.labels:
                 k += 1
                 for metric, i in self.labels[label]:
-                    for j, n in self.get_results_iter(i, metric, results):
-                        name = self.pmconfig.insts[i][1][j] if not self.dynamic_header else str(n[1])
-                        write_line(metric, k, i, name)
+                    if self.fixed_header:
+                        write_line(metric, k, i, None)
+                    else:
+                        for j, n in self.get_results_iter(i, metric, results):
+                            name = self.pmconfig.insts[i][1][j] if not self.dynamic_header else str(n[1])
+                            write_line(metric, k, i, name)
 
         self.writer.write("\n")
         names = ["", self.delimiter] # no timestamp on header line
@@ -782,7 +835,6 @@ class PMReporter(object):
             if self.context.type != PM_CONTEXT_ARCHIVE:
                 self.writer.write(" (Ctrl-C to stop)")
             self.writer.write("\n")
-        return
 
     def write_header_csv(self, results=()):
         """ Write info header for CSV output """
@@ -834,7 +886,8 @@ class PMReporter(object):
             labels.append(self.metrics[metric][0])
         for i, metric in enumerate(self.metrics):
             if self.colxrow is not None:
-                if self.metrics[metric][0] in labels:
+                if self.metrics[metric][0] in labels or \
+                   (self.dynamic_header and results and not results[metric]):
                     continue
                 add_header_items(metric, None)
                 continue
@@ -857,13 +910,23 @@ class PMReporter(object):
             self.writer.write(self.format.format(*units) + "\n")
 
     def write_archive(self, timestamp):
-        """ Write an archive record """
+        """ Write archive record """
         if timestamp is None:
             # Complete and close
             self.pmi.pmiEnd()
             self.pmi = None
             return
 
+        def record_metric_info(metric, i):
+            """ Helper to record metric info """
+            self.recorded[metric] = []
+            self.pmi.pmiAddMetric(metric,
+                                  self.pmconfig.pmids[i],
+                                  self.pmconfig.descs[i].contents.type,
+                                  self.pmconfig.descs[i].contents.indom,
+                                  self.pmconfig.descs[i].contents.sem,
+                                  self.pmconfig.descs[i].contents.units)
+
         if self.pmi is None:
             # Create a new archive
             self.pmi = pmi.pmiLogImport(self.outfile)
@@ -873,18 +936,17 @@ class PMReporter(object):
                 self.pmi.pmiSetHostname(self.context.pmGetArchiveLabel().hostname)
             self.pmi.pmiSetTimezone(self.context.get_current_tz(self.opts))
             for i, metric in enumerate(self.metrics):
-                self.recorded[metric] = []
-                self.pmi.pmiAddMetric(metric,
-                                      self.pmconfig.pmids[i],
-                                      self.pmconfig.descs[i].contents.type,
-                                      self.pmconfig.descs[i].contents.indom,
-                                      self.pmconfig.descs[i].contents.sem,
-                                      self.pmconfig.descs[i].contents.units)
+                record_metric_info(metric, i)
 
         # Add current values
         data = 0
-        results = self.pmconfig.get_sorted_results()
+        # NB. We use valid_only=False to make sure that for every metric
+        # requested their metadata will be recorded in the archive even
+        # if their values are not available for whatever reason.
+        results = self.pmconfig.get_ranked_results(valid_only=False)
         for i, metric in enumerate(results):
+            if metric not in self.recorded:
+                record_metric_info(metric, i)
             for inst, name, value in results[metric]:
                 if inst != PM_IN_NULL and inst not in self.recorded[metric]:
                     try:
@@ -897,9 +959,9 @@ class PMReporter(object):
                     def lookup_inst_index(mres, instance):
                         """ Helper to lookup instance index """
                         index = -1
-                        for inst, _, _ in mres:
+                        for ins, _, _ in mres:
                             index += 1
-                            if inst == instance:
+                            if ins == instance:
                                 return index
                         return -1
                     index = lookup_inst_index(self.prev_res[metric], inst)
@@ -923,10 +985,11 @@ class PMReporter(object):
         """ Update dynamic header as needed """
         if self.rank:
             for metric in results:
-                for i in results[metric]:
-                    results[metric] = sorted(results[metric], key=lambda x: x[0])
+                results[metric] = sorted(results[metric], key=lambda x: x[0])
         insts = [i[0] for metric in results for i in results[metric]]
-        if insts and (self.repeat_header == self.lines or insts != self.prev_insts):
+        if self.fixed_header:
+            self.prepare_stdout_colxrow(results)
+        elif insts and (self.repeat_header == self.lines or insts != self.prev_insts):
             if self.output == OUTPUT_CSV:
                 self.write_header_csv(results)
             if self.output == OUTPUT_STDOUT:
@@ -936,7 +999,8 @@ class PMReporter(object):
                     self.prepare_stdout_colxrow(results)
                 self.write_header_stdout(self.prev_insts is not None, results)
             self.lines = 0
-        self.lines += 1
+        if not self.fixed_header:
+            self.lines += 1
         if not insts and line:
             self.format = "{0:}{1}{2:>" + str(self.width) + "}"
             line.extend([NO_VAL, self.delimiter])
@@ -989,7 +1053,7 @@ class PMReporter(object):
         if self.extcsv:
             line += self.csv_tz
 
-        results = self.pmconfig.get_sorted_results()
+        results = self.pmconfig.get_ranked_results()
 
         if self.dynamic_header:
             self.dynamic_header_update(results)
@@ -999,7 +1063,7 @@ class PMReporter(object):
             for inst, _, value in results[metric]:
                 res[metric + "+" + str(inst)] = value
 
-        # Add corresponding values for each column in the static header
+        # Add corresponding values for each header column
         for i, metric in enumerate(self.metrics):
             fmt = "." + str(self.metrics[metric][6]) + "f"
             for j, n in self.get_results_iter(i, metric, results):
@@ -1023,7 +1087,7 @@ class PMReporter(object):
         self.writer.write(line + "\n")
 
     def format_stdout_value(self, value, width, precision, fmt, k):
-        """ Format a value for stdout output """
+        """ Format value for stdout output """
         if isinstance(value, (int, long)):
             if len(str(value)) > width:
                 value = pmconfig.TRUNC
@@ -1038,7 +1102,7 @@ class PMReporter(object):
                 value = pmconfig.TRUNC
             elif s + 2 > width:
                 fmt[k] = "{X:" + str(width) + "d}"
-                value = int(value) # pylint: disable=redefined-variable-type
+                value = int(value)
             else:
                 c = precision
                 for _ in reversed(range(c+1)):
@@ -1058,14 +1122,14 @@ class PMReporter(object):
         return value
 
     def write_stdout(self, timestamp):
-        """ Write a line to stdout """
+        """ Write line to stdout """
         if self.colxrow is None:
             self.write_stdout_std(timestamp)
         else:
             self.write_stdout_colxrow(timestamp)
 
     def write_stdout_std(self, timestamp):
-        """ Write a line to standard formatted stdout """
+        """ Write line to standard formatted stdout """
         if timestamp is None:
             # Silent goodbye
             return
@@ -1077,7 +1141,7 @@ class PMReporter(object):
             line.append(timestamp)
         line.append(self.delimiter)
 
-        results = self.pmconfig.get_sorted_results()
+        results = self.pmconfig.get_ranked_results()
 
         if self.dynamic_header:
             self.dynamic_header_update(results, line)
@@ -1090,7 +1154,7 @@ class PMReporter(object):
             for inst, _, value in results[metric]:
                 res[metric + "+" + str(inst)] = value
 
-        # Add corresponding values for each column in the static header
+        # Add corresponding values for each header column
         k = 0
         for i, metric in enumerate(self.metrics):
             for j, n in self.get_results_iter(i, metric, results):
@@ -1117,7 +1181,7 @@ class PMReporter(object):
         self.writer.write(nfmt.format(*line) + "\n")
 
     def write_stdout_colxrow(self, timestamp):
-        """ Write a line to columns and rows swapped stdout """
+        """ Write line to columns and rows swapped stdout """
         if timestamp is None:
             # Silent goodbye
             return
@@ -1125,19 +1189,30 @@ class PMReporter(object):
         # Avoid per-line I/O
         output = ""
 
-        results = self.pmconfig.get_sorted_results()
+        results = self.pmconfig.get_ranked_results()
 
         res = {}
         for i, metric in enumerate(results):
             for inst, name, value in results[metric]:
-                if not self.dynamic_header:
+                if self.static_header:
                     res[metric + "+" + str(inst)] = value
                 else:
                     res[metric + "+" + str(name)] = value
 
-        if self.dynamic_header:
+        if not self.static_header:
             self.dynamic_header_update(results)
 
+        if self.sort_metric:
+            found_insts = self.found_insts
+            self.found_insts = []
+            for sort_metric in self.sort_metric.split(","):
+                revs = sort_metric[:1] != "-"
+                sort_metric = sort_metric if revs else sort_metric[1:]
+                for r in sorted(results[sort_metric], key=lambda x: x[2], reverse=revs):
+                    if r[1] not in self.found_insts:
+                        self.found_insts.append(r[1])
+            self.found_insts.extend([i for i in found_insts if i not in self.found_insts])
+
         # We need to construct each line independently
         for instance in self.found_insts:
             # Split on dummies
@@ -1168,18 +1243,18 @@ class PMReporter(object):
                 for metric, i in self.labels[label]:
                     if found:
                         break
-                    insts = self.pmconfig.insts[i][1] if not self.dynamic_header else self.found_insts
+                    insts = self.pmconfig.insts[i][1] if self.static_header else self.found_insts
                     if label == self.metrics[metric][0] and instance in insts:
                         found = 1
                         try:
-                            if not self.dynamic_header:
+                            if self.static_header:
                                 ref = self.pmconfig.insts[i][0][self.pmconfig.insts[i][1].index(instance)]
                             else:
                                 ref = instance
                             value = res[metric + "+" + str(ref)]
                             value = self.format_stdout_value(value, self.metrics[metric][4], self.metrics[metric][6], fmt, k)
                         except Exception:
-                            value = NO_VAL if not self.dynamic_header else NO_INST
+                            value = NO_VAL if self.static_header else NO_INST
 
                         line.extend([value, self.delimiter])
                         k += 1
@@ -1212,8 +1287,13 @@ class PMReporter(object):
 
         if not output:
             line = [""] if self.timestamp == 0 else [timestamp]
-            self.format = "{0:}{1}{2:>" + str(self.width) + "}"
-            line.extend([self.delimiter, NO_VAL, self.delimiter])
+            if self.dynamic_header:
+                self.format = "{0:}{1}{2:>" + str(len(self.colxrow)) + "}"
+                line.extend([self.delimiter, NO_VAL, self.delimiter])
+            else:
+                line.extend([self.delimiter, NO_VAL])
+                for _ in range(len(self.metrics)):
+                    line.extend([self.delimiter, NO_INST])
             output = self.format.format(*line) + "\n"
 
         self.writer.write(output)
@@ -1266,7 +1346,7 @@ class PMReporter(object):
                     self.writer.write(output.format(*alt_line) + "\n")
             return
 
-        results = self.pmconfig.get_sorted_results()
+        results = self.pmconfig.get_ranked_results()
 
         if self.prev_insts is None:
             for i, metric in enumerate(results):
@@ -1274,18 +1354,18 @@ class PMReporter(object):
                     self.all_ranked[metric] = results[metric]
             self.prev_insts = []
 
-        revs = True if self.rank > 0 else False
+        revs = self.rank > 0
 
         for i, metric in enumerate(results):
             if self.pmconfig.descs[i].contents.type == PM_TYPE_STRING:
                 continue
             rank = abs(self.rank) if self.pmconfig.descs[i].contents.indom != PM_IN_NULL else 1
             c, r, t = (0, [], [])
-            for i in sorted(results[metric] + self.all_ranked[metric], key=lambda value: value[2], reverse=revs):
-                if i[0] not in t and c < rank:
+            for j in sorted(results[metric] + self.all_ranked[metric], key=lambda x: x[2], reverse=revs):
+                if j[0] not in t and c < rank:
                     c += 1
-                    r.append(i)
-                    t.append(i[0])
+                    r.append(j)
+                    t.append(j[0])
             self.all_ranked[metric] = r
 
     def finalize(self):
@@ -1293,12 +1373,12 @@ class PMReporter(object):
         if self.writer:
             try:
                 self.writer.flush()
-            except socket.error as error:
+            except IOError as error:
                 if error.errno != errno.EPIPE:
                     raise error
             try:
                 self.writer.close()
-            except: # pylint: disable=bare-except
+            except Exception:
                 pass
             self.writer = None
         if self.pmi:
@@ -1312,9 +1392,11 @@ if __name__ == '__main__':
         P.validate_config()
         P.execute()
         P.finalize()
-
     except pmapi.pmErr as error:
-        sys.stderr.write('%s: %s\n' % (error.progname(), error.message()))
+        sys.stderr.write("%s: %s" % (error.progname(), error.message()))
+        if error.message() == "Connection refused":
+            sys.stderr.write("; is pmcd running?")
+        sys.stderr.write("\n")
         sys.exit(1)
     except pmapi.pmUsageErr as usage:
         usage.message()
