@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2017-2018 Red Hat.
+ * Copyright (c) 2017-2019 Red Hat.
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -12,12 +12,8 @@
  * for more details.
  */
 
-#include "series.h"
-#include "pmapi.h"
-#include "sds.h"
-
-unsigned int port = 6379;
-sds hostname = "localhost";
+#include "pmwebapi.h"
+#include <uv.h>
 
 typedef enum series_flags {
     PMSERIES_COLOUR	= (1<<0),	/* report in colour if possible */
@@ -25,16 +21,16 @@ typedef enum series_flags {
     PMSERIES_FULLINDOM	= (1<<2),	/* report with pminfo(1) -I info */
     PMSERIES_FULLPMID	= (1<<3),	/* report with pminfo(1) -M info */
     PMSERIES_SERIESID	= (1<<4),	/* report with pminfo(1) -s info */
-    PMSERIES_SOURCEID	= (1<<5),	/* report with pminfo(1) -S info */
-    PMSERIES_NEED_EOL	= (1<<6),	/* need to eol-terminate output */
-    PMSERIES_NEED_COMMA	= (1<<7),	/* need comma line separation */
-    PMSERIES_INSTLABELS	= (1<<8),	/* labels by instance identifier */
-    PMSERIES_ONLY_NAMES	= (1<<9),	/* report on label names only */
-    PMSERIES_NEED_DESCS	= (1<<10),	/* output requires descs lookup */
-    PMSERIES_NEED_INSTS	= (1<<11),	/* output requires insts lookup */
+    PMSERIES_NEED_EOL	= (1<<5),	/* need to eol-terminate output */
+    PMSERIES_NEED_COMMA	= (1<<6),	/* need comma line separation */
+    PMSERIES_INSTLABELS	= (1<<7),	/* labels by instance identifier */
+    PMSERIES_ONLY_NAMES	= (1<<8),	/* report on label names only */
+    PMSERIES_NEED_DESCS	= (1<<9),	/* output requires descs lookup */
+    PMSERIES_NEED_INSTS	= (1<<10),	/* output requires insts lookup */
+    PMSERIES_NEED_RESET	= (1<<11),	/* need to reset for next series */
 
     PMSERIES_OPT_ALL	= (1<<16),	/* -a, --all option */
-    PMSERIES_OPT_SOURCE = (1<<17),	/* -c, --context option */
+    PMSERIES_OPT_SOURCE = (1<<17),	/* -S, --source option */
     PMSERIES_OPT_DESC	= (1<<18),	/* -d, --desc option */
     PMSERIES_OPT_INSTS	= (1<<19),	/* -i, --instances option */
     PMSERIES_OPT_LABELS	= (1<<20),	/* -l, --labels option */
@@ -46,6 +42,23 @@ typedef enum series_flags {
 #define PMSERIES_META_OPTS	(PMSERIES_OPT_DESC | PMSERIES_OPT_INSTS | \
 				 PMSERIES_OPT_LABELS | PMSERIES_OPT_METRIC)
 
+struct series_data;
+typedef void (*series_callback)(struct series_data *, void *);
+
+typedef struct series_entry {
+    series_callback	func;
+    void		*arg;
+    struct series_entry	*next;
+} series_entry;
+
+typedef struct series_command {
+    int			nseries;
+    int			nsource;
+    sds			*series;
+    sds			*source;
+    sds			pattern;	/* glob pattern for string matches */
+} series_command;
+
 typedef struct series_label {
     sds			name;
     sds			value;
@@ -60,8 +73,14 @@ typedef struct series_inst {
 } series_inst;
 
 typedef struct series_data {
-    int			status;		/* command exit status */
+    sds			query;
+    uv_loop_t		*loop;
+    pmSeriesSettings	settings;
+    series_command	args;		/* detailed command line arguments */
+    series_entry	*head;		/* list of function callbacks */
+    series_entry	*next;		/* next function callback to issue */
     series_flags	flags;		/* flags affecting reporting */
+    int			status;		/* command exit status */
 
     pmSID		series;		/* current time series */
     pmSID		source;		/* current time series source */
@@ -75,16 +94,23 @@ typedef struct series_data {
     pmSID		*iseries;	/* series identifiers for instances */
 } series_data;
 
-#define series_data_endtopic(dp) 	((dp)->flags &= ~PMSERIES_NEED_COMMA)
-#define series_data_endline(dp) 	((dp)->flags &= ~PMSERIES_NEED_EOL)
+static void on_series_done(int, void *);
 
-static void
-series_data_init(series_data *dp, series_flags flags)
+static series_data *
+series_data_init(series_flags flags, sds query)
 {
-    memset(dp, 0, sizeof(series_data));
-    dp->series = sdsnewlen("", 40);
-    dp->source = sdsnewlen("", 40);
+    series_data		*dp = calloc(1, sizeof(series_data));
+
+    if (dp == NULL) {
+	fprintf(stderr, "%s: out of memory allocating series data\n",
+			pmGetProgname());
+	exit(127);
+    }
+    dp->query = query;
     dp->flags = flags;
+    dp->series = sdsempty();
+    dp->source = sdsempty();
+    return dp;
 }
 
 static void
@@ -119,47 +145,47 @@ series_add_inst(series_data *dp, pmSID series, sds instid, sds instname)
     }
 }
 
+static void
+labels_free(series_label *labels, unsigned int nlabels)
+{
+    series_label	*lp;
+    unsigned int	i;
+
+    for (i = 0; i < nlabels; i++) {
+	lp = &labels[i];
+	sdsfree(lp->name);
+	sdsfree(lp->value);
+    }
+}
+
+static void
+series_del_labels(series_data *dp)
+{
+    labels_free(dp->labels, dp->nlabels);
+    dp->labels = NULL;
+    dp->nlabels = 0;
+}
+
 static void
 series_del_insts(series_data *dp)
 {
     series_inst		*ip;
-    int			i;
+    unsigned int	i;
 
     for (i = 0; i < dp->ninsts; i++) {
 	ip = &dp->insts[i];
+	labels_free(ip->labels, ip->nlabels);
 	sdsfree(ip->series);
 	sdsfree(ip->instid);
 	sdsfree(ip->name);
     }
-    if (dp->insts)
+    if (dp->insts) {
 	free(dp->insts);
+	dp->insts = NULL;
+    }
     dp->ninsts = 0;
 }
 
-static void
-series_data_free(series_data *dp)
-{
-    sdsfree(dp->series);
-    sdsfree(dp->source);
-    if (dp->type)
-	sdsfree(dp->type);
-    series_del_insts(dp);
-}
-
-static int
-series_split(sds string, pmSID **series)
-{
-    size_t		length;
-    int			nseries = 0;
-
-    if (!string || !sdslen(string))
-	return 0;
-    length = strlen(string);
-    if ((*series = sdssplitlen(string, length, ",", 1, &nseries)) == NULL)
-	return -ENOMEM;
-    return nseries;
-}
-
 static void
 series_free(int nseries, pmSID *series)
 {
@@ -170,6 +196,52 @@ series_free(int nseries, pmSID *series)
     }
 }
 
+static void
+series_data_reset(series_data *dp)
+{
+    sdsclear(dp->series);
+    sdsclear(dp->source);
+
+    if (dp->type)
+	sdsfree(dp->type);
+    dp->type = NULL;
+
+    dp->flags &= ~PMSERIES_INSTLABELS;
+    series_del_labels(dp);
+    series_del_insts(dp);
+}
+
+static void
+series_data_free(series_data *dp)
+{
+    if (dp->args.nsource)
+	series_free(dp->args.nsource, dp->args.source);
+    if (dp->args.nseries)
+	series_free(dp->args.nseries, dp->args.series);
+    if (dp->args.pattern)
+	sdsfree(dp->args.pattern);
+
+    series_data_reset(dp);
+
+    sdsfree(dp->series);
+    sdsfree(dp->source);
+    sdsfree(dp->query);
+}
+
+static int
+series_split(sds string, pmSID **series)
+{
+    size_t		length;
+    int			nseries = 0;
+
+    if (!string || !sdslen(string))
+	return 0;
+    length = sdslen(string);
+    if ((*series = sdssplitlen(string, length, ",", 1, &nseries)) == NULL)
+	return -ENOMEM;
+    return nseries;
+}
+
 static series_inst *
 series_get_inst(series_data *dp, sds series)
 {
@@ -187,17 +259,18 @@ series_get_inst(series_data *dp, sds series)
 static int
 series_next(series_data *dp, sds sid)
 {
-    if (strncmp(dp->series, sid, sdslen(sid)) != 0) {
-	dp->flags &= ~PMSERIES_NEED_COMMA;
+    if (sdscmp(dp->series, sid) != 0) {
+	dp->flags &= ~(PMSERIES_NEED_COMMA|PMSERIES_NEED_RESET);
 	if (dp->flags & PMSERIES_NEED_EOL) {
 	    dp->flags &= ~PMSERIES_NEED_EOL;
 	    putc('\n', stdout);
 	}
-	dp->series = sdscpylen(dp->series, sid, sdslen(sid));
+	dp->series = sdscpy(dp->series, sid);
 	if (dp->source)
 	    sdsclear(dp->source);
 	if (dp->type)
 	    sdsclear(dp->type);
+	series_del_labels(dp);
 	series_del_insts(dp);
 	return 1;
     }
@@ -205,13 +278,16 @@ series_next(series_data *dp, sds sid)
 }
 
 static void
-on_series_info(pmloglevel level, sds message, void *arg)
+on_series_info(pmLogLevel level, sds message, void *arg)
 {
     series_data		*dp = (series_data *)arg;
     int			colour = (dp->flags & PMSERIES_COLOUR);
     FILE		*fp = (level == PMLOG_INFO) ? stdout : stderr;
 
-    return pmLogLevelPrint(fp, level, message, colour);
+    if (level >= PMLOG_ERROR)
+	dp->status = 1;	/* exit pmseries with error */
+    if (level >= PMLOG_INFO || pmDebugOptions.series)
+	pmLogLevelPrint(fp, level, message, colour);
 }
 
 static const char *
@@ -244,15 +320,15 @@ series_type_phrase(const char *type_word)
     return "???";
 }
 
-static int
-series_load(pmSeriesSettings *settings, sds query, series_flags flags)
+static void
+series_load(series_data *dp)
 {
-    series_data		data;
-    pmflags		meta = flags & PMSERIES_FAST? PMFLAG_METADATA : 0;
+    pmSeriesFlags	meta = dp->flags & PMSERIES_FAST ?
+				PM_SERIES_FLAG_METADATA : 0;
+    int			sts;
 
-    series_data_init(&data, flags);
-    pmSeriesLoad(settings, query, meta, (void *)&data);
-    return data.status;
+    if ((sts = pmSeriesLoad(&dp->settings, dp->query, meta, dp)) < 0)
+	on_series_done(sts, dp);
 }
 
 static int
@@ -268,8 +344,8 @@ on_series_match(pmSID sid, void *arg)
 static int
 on_series_value(pmSID sid, pmSeriesValue *value, void *arg)
 {
-    series_inst		*ip;
     series_data		*dp = (series_data *)arg;
+    series_inst		*ip;
     sds			timestamp, series, data;
     int			need_free = 1;
 
@@ -301,15 +377,15 @@ on_series_value(pmSID sid, pmSeriesValue *value, void *arg)
     return 0;
 }
 
-static int
-series_query(pmSeriesSettings *settings, sds query, series_flags flags)
+static void
+series_query(series_data *dp)
 {
-    series_data		data;
-    pmflags		meta = flags & PMSERIES_FAST? PMFLAG_METADATA : 0;
+    pmSeriesFlags	meta = dp->flags & PMSERIES_FAST ?
+		   		PM_SERIES_FLAG_METADATA : 0;
+    int			sts;
 
-    series_data_init(&data, flags);
-    pmSeriesQuery(settings, query, meta, (void *)&data);
-    return data.status;
+    if ((sts = pmSeriesQuery(&dp->settings, dp->query, meta, dp)) < 0)
+	on_series_done(sts, dp);
 }
 
 static int
@@ -371,7 +447,7 @@ on_series_desc(pmSID series, pmSeriesDesc *desc, void *arg)
 	    printf(" (%s)", semantics);
 	printf("  Units: %s\n", *units == '\0' ? "none" : units);
     }
-    if (dp->flags & PMSERIES_SOURCEID)
+    if (dp->flags & PMSERIES_SERIESID)
 	printf("    Source: %s\n", source);
     dp->flags &= ~PMSERIES_NEED_EOL;
 
@@ -428,13 +504,13 @@ series_instance_compare(const void *a, const void *b)
 }
 
 static void
-series_instance_names(void *arg)
+series_instance_names(series_data *dp, void *arg)
 {
-    series_data		*dp = (series_data *)arg;
     series_inst		*ip;
     pmSID		*isp = dp->iseries;
     int			i;
 
+    (void)arg;
     qsort(dp->insts, dp->ninsts, sizeof(series_inst), series_instance_compare);
     for (i = 0; i < dp->ninsts; i++) {
 	ip = &dp->insts[i];
@@ -443,6 +519,7 @@ series_instance_names(void *arg)
 		    ip->instid, ip->name, ip->series);
 	isp[i] = ip->series;
     }
+    on_series_done(0, dp);
 }
 
 static int
@@ -464,6 +541,8 @@ series_labels_sort(sds s, unsigned int nlabels, series_label *labels)
 
     s = sdscpylen(s, "{", 1);
     for (i = 0; i < nlabels; i++) {
+	if (i > 0)
+	    s = sdscat(s, ",");
 	lp = &labels[i];
 	s = sdscatfmt(s, "\"%S\":%S", lp->name, lp->value);
     }
@@ -471,23 +550,23 @@ series_labels_sort(sds s, unsigned int nlabels, series_label *labels)
 }
 
 static void
-series_metric_labels(void *arg)
+series_metric_labels(series_data *dp, void *arg)
 {
-    series_data		*dp = (series_data *)arg;
     sds			labels;
 
+    (void)arg;
     if (!(dp->flags & PMSERIES_ONLY_NAMES)) {
 	labels = series_labels_sort(sdsempty(), dp->nlabels, dp->labels);
 	if (sdslen(labels) > 2)
 	    printf("    labels %s\n", labels);
 	sdsfree(labels);
     }
+    on_series_done(0, dp);
 }
 
 static void
-series_instance_labels(void *arg)
+series_instance_labels(series_data *dp, void *arg)
 {
-    series_data		*dp = (series_data *)arg;
     series_inst		*ip = NULL;
     sds			labels;
     unsigned int	i;
@@ -502,6 +581,7 @@ series_instance_labels(void *arg)
 	}
 	sdsfree(labels);
     }
+    on_series_done(0, dp);
 }
 
 static void
@@ -509,14 +589,15 @@ series_add_labels(sds name, sds value,
 		unsigned int *nlabelsp, series_label **labelsp)
 {
     unsigned int	nlabels = *nlabelsp;
-    series_label	*lp = *labelsp;
+    series_label	*p, *lp = *labelsp;
     size_t		bytes;
 
     bytes = sizeof(series_label) * (nlabels + 1);
-    if ((lp = realloc(lp, bytes)) != NULL) {
-	lp += nlabels;
-	lp->name = sdsdup(name);
-	lp->value = sdsdup(value);
+    if ((p = realloc(lp, bytes)) != NULL) {
+	lp = p;
+	p += nlabels;
+	p->name = sdsdup(name);
+	p->value = sdsdup(value);
 
 	*labelsp = lp;
 	*nlabelsp = nlabels + 1;
@@ -611,112 +692,316 @@ on_series_context(pmSID source, sds name, void *arg)
     return 0;
 }
 
-static int
-series_source(pmSeriesSettings *settings, sds query, series_flags flags)
+static void
+series_source(series_data *dp)
 {
     int			nsources, sts;
     char		msg[PM_MAXERRMSGLEN];
     pmSID		*sources = NULL;
-    series_data		data;
 
-    if ((nsources = sts = series_split(query, &sources)) < 0) {
+    if ((nsources = sts = series_split(dp->query, &sources)) < 0) {
 	fprintf(stderr, "%s: cannot find source identifiers in '%s': %s\n",
-		pmGetProgname(), query, pmErrStr_r(sts, msg, sizeof(msg)));
-	return 2;
+		pmGetProgname(), dp->query, pmErrStr_r(sts, msg, sizeof(msg)));
+    } else {
+	if (nsources) {
+	    dp->args.nsource = nsources;
+	    dp->args.source = sources;
+	} else {
+	    sources = &dp->args.pattern;
+	}
+	if ((sts = pmSeriesSources(&dp->settings, nsources, sources, dp)) < 0)
+	    on_series_done(sts, dp);
     }
-    series_data_init(&data, flags);
-    pmSeriesSources(settings, nsources, sources, (void *)&data);
-    series_free(nsources, sources);
-    return data.status;
 }
 
 /*
  * Finishing up interacting with the library via callbacks
  */
 
-void
+static void
 on_series_done(int sts, void *arg)
 {
+    series_callback	func;
+    series_entry	*entry;
     series_data		*dp = (series_data *)arg;
     char		msg[PM_MAXERRMSGLEN];
 
+    if (dp->flags & PMSERIES_NEED_RESET) {
+	dp->flags &= ~PMSERIES_NEED_RESET;
+	series_data_reset(dp);
+    }
     if (dp->flags & PMSERIES_NEED_EOL) {
 	dp->flags &= ~PMSERIES_NEED_EOL;
 	putc('\n', stdout);
     }
-    if (sts < 0) {
+    if (dp->status == 0 && sts < 0) {
 	fprintf(stderr, "%s: %s\n", pmGetProgname(),
 			pmErrStr_r(sts, msg, sizeof(msg)));
+	dp->status = 1;
+    }
+
+    if ((entry = dp->next) != NULL) {
+	dp->next = entry->next;
+	func = entry->func;
+	arg = entry->arg;
+	func(dp, arg);
+    } else {
+	pmSeriesClose(&dp->settings.module);
+	series_data_free(dp);
     }
-    dp->status = (sts < 0) ? 1 : 0;
 }
 
-static int
-series_data_report(pmSeriesSettings *settings,
-		int nseries, pmSID series, series_flags flags)
+/* append the given function pointer into the callback chain */
+static void
+series_link_report(series_data *dp, series_callback func, void *arg)
 {
-    series_data		data;
-    int			sts;
+    series_entry	*tail, *entry = calloc(1, sizeof(series_entry));
 
-    series_data_init(&data, flags);
-    if (nseries && series_next(&data, series))
-	printf("\n%s\n", series);
+    if (entry == NULL) {
+	fprintf(stderr, "%s: out of memory allocating callback chain\n",
+			pmGetProgname());
+	exit(127);
+    }
+    entry->func = func;
+    entry->arg = arg;
 
-    if (flags & (PMSERIES_OPT_DESC|PMSERIES_NEED_DESCS)) {
-	pmSeriesDescs(settings, nseries, &series, (void *)&data);
-	series_data_endtopic(&data);
+    if (dp->head == NULL) {
+	dp->head = dp->next = entry;
+    } else {
+	tail = dp->head;
+	while (tail->next != NULL)
+	    tail = tail->next;
+	tail->next = entry;
     }
-    if (flags & PMSERIES_OPT_SOURCE) {
-	pmSeriesSources(settings, 1, &data.source, (void *)&data);
-	series_data_endtopic(&data);
+}
+
+static void
+series_report_header(series_data *dp, void *arg)
+{
+    if (arg != NULL && series_next(dp, (pmSID)arg))
+	printf("\n%s\n", (pmSID)arg);
+    on_series_done(0, dp);
+}
+
+static void
+series_done_report(series_data *dp, void *arg)
+{
+    (void)arg;
+    dp->flags &= ~PMSERIES_NEED_COMMA;
+    on_series_done(0, dp);
+}
+
+static void
+series_report_footer(series_data *dp, void *arg)
+{
+    (void)arg;
+    dp->flags &= ~PMSERIES_NEED_EOL;
+    on_series_done(0, dp);
+}
+
+static void
+series_report_reset(series_data *dp, void *arg)
+{
+    (void)arg;
+    dp->flags |= PMSERIES_NEED_RESET;
+    on_series_done(0, dp);
+}
+
+static void
+series_desc_report(series_data *dp, void *arg)
+{
+    pmSID	sid = (pmSID)arg;
+    int		sts;
+
+    if ((sts = pmSeriesDescs(&dp->settings, sid? 1 : 0, &sid, dp)) < 0)
+	on_series_done(sts, dp);
+}
+
+/* pass series identifier (one count) or pattern (zero count) */
+#define SERIES_PARAMS(dp, arg, count, param) \
+	if (arg) { count = 1; param = arg; } \
+	else { count = 0; param = dp->args.pattern; }
+
+static void
+series_source_report(series_data *dp, void *arg)
+{
+    sds		param;
+    int		sts, count;
+
+    SERIES_PARAMS(dp, arg, count, param);
+    if ((sts = pmSeriesSources(&dp->settings, count, &param, dp)) < 0)
+	on_series_done(sts, dp);
+}
+
+static void
+series_metric_report(series_data *dp, void *arg)
+{
+    sds		param;
+    int		sts, count;
+
+    SERIES_PARAMS(dp, arg, count, param);
+    if ((sts = pmSeriesMetrics(&dp->settings, count, &param, dp)) < 0)
+	on_series_done(sts, dp);
+}
+
+static void
+series_labels_report(series_data *dp, void *arg)
+{
+    sds		param;
+    int		sts, count;
+
+    SERIES_PARAMS(dp, arg, count, param);
+    if ((sts = pmSeriesLabels(&dp->settings, count, &param, dp)) < 0)
+	on_series_done(sts, dp);
+}
+
+static void
+series_instances_report(series_data *dp, void *arg)
+{
+    sds		param;
+    int		sts, count;
+
+    SERIES_PARAMS(dp, arg, count, param);
+    if ((sts = pmSeriesInstances(&dp->settings, count, &param, dp)) < 0)
+	on_series_done(sts, dp);
+}
+
+static void
+series_instlabels_report(series_data *dp, void *arg)
+{
+    int		sts;
+
+    (void)arg;
+    if (dp->ninsts > 0) {
+	dp->flags |= PMSERIES_INSTLABELS;
+	sts = pmSeriesLabels(&dp->settings, dp->ninsts, dp->iseries, dp);
+	if (sts < 0)
+	    on_series_done(sts, dp);
+    } else {
+	/* nothing to do - move on to next command handler */
+	on_series_done(0, dp);
     }
-    if (flags & PMSERIES_OPT_METRIC) {
-	pmSeriesMetrics(settings, nseries, &series, (void *)&data);
-	series_data_endtopic(&data);
+}
+
+/*
+ * Setup the (async) callback chain that will produce the
+ * pmseries output requested based on command line option
+ * list presented.  This may be called iteratively to add
+ * more series than just one, or may be called for *zero*
+ * series in the 'report all metadata' modes.
+ */
+static void
+series_data_report(series_data *dp, int nseries, pmSID series)
+{
+    series_link_report(dp, series_report_header, series);
+
+    if (dp->flags & (PMSERIES_OPT_DESC|PMSERIES_NEED_DESCS)) {
+	series_link_report(dp, series_desc_report, series);
+	series_link_report(dp, series_done_report, series);
     }
-    if (flags & PMSERIES_OPT_LABELS) {
-	pmSeriesLabels(settings, nseries, &series, (void *)&data);
-	series_metric_labels(&data);
-	series_data_endtopic(&data);
+
+    if (dp->flags & PMSERIES_OPT_SOURCE) {
+	series_link_report(dp, series_source_report, series);
+	series_link_report(dp, series_done_report, series);
     }
-    if (flags & (PMSERIES_OPT_INSTS|PMSERIES_NEED_INSTS)) {
-	pmSeriesInstances(settings, nseries, &series, (void *)&data);
-	series_instance_names(&data);
-	series_data_endtopic(&data);
+
+    if (dp->flags & PMSERIES_OPT_METRIC) {
+	series_link_report(dp, series_metric_report, series);
+	series_link_report(dp, series_done_report, series);
     }
+
+    if (dp->flags & PMSERIES_OPT_LABELS) {
+	series_link_report(dp, series_labels_report, series);
+	series_link_report(dp, series_metric_labels, series);
+	series_link_report(dp, series_done_report, series);
+    }
+
+    if (dp->flags & (PMSERIES_OPT_INSTS|PMSERIES_NEED_INSTS)) {
+	series_link_report(dp, series_instances_report, series);
+	series_link_report(dp, series_instance_names, series);
+	series_link_report(dp, series_done_report, series);
+    }
+
     /* report per-instance label information */
-    if ((flags & PMSERIES_OPT_LABELS) && nseries != 0) {
-	data.flags |= PMSERIES_INSTLABELS;
-	pmSeriesLabels(settings, data.ninsts, data.iseries, (void *)&data);
-	series_instance_labels(&data);
-	series_data_endtopic(&data);
+    if ((dp->flags & PMSERIES_OPT_LABELS) && nseries != 0) {
+	series_link_report(dp, series_instlabels_report, series);
+	series_link_report(dp, series_instance_labels, series);
+	series_link_report(dp, series_done_report, series);
     }
-    series_data_endline(&data);
 
-    sts = data.status;
-    series_data_free(&data);
-    return sts;
+    series_link_report(dp, series_report_footer, series);
 }
 
-static int
-series_report(pmSeriesSettings *settings, sds query, series_flags flags)
+static void
+series_report(series_data *dp)
 {
-    int			nseries, sts, rc, i;
+    int			nseries, sts, i;
     char		msg[PM_MAXERRMSGLEN];
     pmSID		*series = NULL;
+    series_entry	*entry;
 
-    if ((nseries = sts = series_split(query, &series)) < 0) {
+    if ((nseries = sts = series_split(dp->query, &series)) < 0) {
 	fprintf(stderr, "%s: no series identifiers in string '%s': %s\n",
-		pmGetProgname(), query, pmErrStr_r(sts, msg, sizeof(msg)));
-	return 2;
+		pmGetProgname(), dp->query, pmErrStr_r(sts, msg, sizeof(msg)));
+    } else {
+	dp->args.nseries = nseries;
+	dp->args.series = series;
+
+	if (nseries == 0)	/* report all names, instances, labels, ... */
+	    series_data_report(dp, 0, NULL);
+	for (i = 0; i < nseries; i++) {
+	    series_data_report(dp, 1, series[i]);
+	    series_link_report(dp, series_report_reset, series);
+	}
+	entry = dp->head;
+	entry->func(dp, entry->arg);
     }
-    for (i = sts = 0; i < nseries; i++)
-	if ((rc = series_data_report(settings, 1, series[i], flags)) < 0)
-	    sts = rc;
-    if (nseries == 0)	/* report all names, instances, labels, ... */
-	sts = series_data_report(settings, 0, NULL, flags);
-    series_free(nseries, series);
-    return sts ? 1 : 0;
+}
+
+static void
+on_series_setup(void *arg)
+{
+    series_data		*dp = (series_data *)arg;
+    series_flags	flags = dp->flags;
+    sds			msg;
+
+    if (pmDebugOptions.series) {
+	msg = sdsnew("Connection established");
+	on_series_info(PMLOG_DEBUG, msg, arg);
+	sdsfree(msg);
+    }
+
+    if (flags & PMSERIES_OPT_LOAD)
+	series_load(dp);
+    else if (flags & PMSERIES_OPT_QUERY)
+	series_query(dp);
+    else if ((flags & PMSERIES_OPT_SOURCE) && !(flags & PMSERIES_META_OPTS))
+	series_source(dp);
+    else
+	series_report(dp);
+}
+
+static void
+pmseries_request(uv_timer_t *arg)
+{
+    uv_handle_t		*handle = (uv_handle_t *)arg;
+    series_data		*dp = (series_data *)handle->data;
+
+    pmSeriesSetup(&dp->settings.module, dp);
+}
+
+static int
+pmseries_execute(series_data *dp)
+{
+    uv_loop_t		*loop = dp->loop;
+    uv_timer_t		request;
+    uv_handle_t		*handle = (uv_handle_t *)&request;
+
+    handle->data = (void *)dp;
+    uv_timer_init(loop, &request);
+    uv_timer_start(&request, pmseries_request, 0, 0);
+    uv_run(loop, UV_RUN_DEFAULT);
+    return dp->status;
 }
 
 static int
@@ -725,6 +1010,7 @@ pmseries_overrides(int opt, pmOptions *opts)
     switch (opt) {
     case 'a':
     case 'h':
+    case 'g':
     case 'L':
     case 's':
     case 'S':
@@ -735,49 +1021,36 @@ pmseries_overrides(int opt, pmOptions *opts)
     return 0;
 }
 
-static pmSeriesSettings settings = {
-    .on_match		= on_series_match,
-    .on_desc		= on_series_desc,
-    .on_inst		= on_series_inst,
-    .on_labelmap	= on_series_labelmap,
-    .on_instance	= on_series_instance,
-    .on_context		= on_series_context,
-    .on_metric		= on_series_metric,
-    .on_value		= on_series_value,
-    .on_label		= on_series_label,
-    .on_info		= on_series_info,
-    .on_done		= on_series_done,
-    .hostspec           = "localhost:6379",
-};
-
 static pmLongOptions longopts[] = {
+    PMAPI_OPTIONS_HEADER("Connection Options"),
+    { "config", 1, 'c', "FILE", "configuration file path"},
+    { "host", 1, 'h', "HOST", "connect to Redis using given host name" },
+    { "port", 1, 'p', "PORT", "connect to Redis using given TCP/IP port" },
     PMAPI_OPTIONS_HEADER("General Options"),
-    { "all", 0, 'a', 0, "report all metadata (-dilmsS) for time series" },
-    { "contexts", 0, 'c', 0, "report context names for a time series source" },
-    { "desc", 0, 'd', 0, "metric descriptor for time series" },
-    { "instances", 0, 'i', 0, "instance identifiers for time series" },
-    { "labels", 0, 'l', 0, "list all labels for time series" },
     { "load", 0, 'L', 0, "load time series values and metadata" },
-    { "metrics", 0, 'm', 0, "metric names for time series" },
     { "query", 0, 'q', 0, "perform a time series query (default)" },
-    { "port", 1, 'p', "N", "Connect to Redis instance on this TCP/IP port" },
-    { "host", 1, 'h', "HOST", "Connect to Redis instance using host specification" },
-    PMAPI_OPTIONS_HEADER("Reporting Options"),
-    PMOPT_DEBUG,
-    { "fast", 0, 'F', 0, "query or load series metadata, not values" },
-    { "fullpmid", 0, 'M', 0, "print PMID in verbose format" },
-    { "fullindom", 0, 'I', 0, "print InDom in verbose format" },
-    { "names", 0, 'n', 0, "print label names only, not values" },
-    { "source", 0, 'S', 0, "print the source for each time series" },
-    { "series", 0, 's', 0, "print the series for each instance" },
     PMOPT_VERSION,
+    PMOPT_DEBUG,
     PMOPT_HELP,
+    PMAPI_OPTIONS_HEADER("Reporting Options"),
+    { "all", 0, 'a', 0, "report all metadata (-dilms) for time series" },
+    { "desc", 0, 'd', 0, "metric descriptor for time series" },
+    { "fullindom", 0, 'I', 0, "print InDom in verbose format" },
+    { "instances", 0, 'i', 0, "report names for time series instances" },
+    { "fast", 0, 'F', 0, "query or load series metadata, not values" },
+    { "glob", 1, 'g', "PATTERN", "glob pattern to restrict matches" },
+    { "labels", 0, 'l', 0, "list all labels for time series" },
+    { "fullpmid", 0, 'M', 0, "print PMID in verbose format" },
+    { "metrics", 0, 'm', 0, "report names for time series metrics" },
+    { "names", 0, 'n', 0, "print label names only, not values" },
+    { "sources", 0, 'S', 0, "report names for time series sources" },
+    { "series", 0, 's', 0, "print series ID for metrics, instances and sources" },
     PMAPI_OPTIONS_END
 };
 
 static pmOptions opts = {
     .flags = PM_OPTFLAG_BOUNDARIES,
-    .short_options = "acdD:Fh:iIlLmMnqp:sSV?",
+    .short_options = "ac:dD:Fg:h:iIlLmMnqp:sSV?",
     .long_options = longopts,
     .short_usage = "[options] [query ... | series ... | source ...]",
     .override = pmseries_overrides,
@@ -786,23 +1059,26 @@ static pmOptions opts = {
 int
 main(int argc, char *argv[])
 {
-    sds			query;
+    sds			option, query, match = NULL;
     int			c, sts;
-    char		*hostname = "localhost";
     const char		*split = ",";
     const char		*space = " ";
-    unsigned int	port = 6379;
+    const char		*inifile = NULL;
+    const char		*redis_host = NULL;
+    unsigned int	redis_port = 6379;	/* default Redis port */
+    struct dict		*config;
     series_flags	flags = 0;
+    series_data		*dp;
 
     while ((c = pmGetOptions(argc, argv, &opts)) != EOF) {
 	switch (c) {
 
 	case 'a':	/* command line contains series identifiers */
-	    flags |= (PMSERIES_OPT_ALL | PMSERIES_SOURCEID | PMSERIES_SERIESID);
+	    flags |= (PMSERIES_OPT_ALL | PMSERIES_SERIESID);
 	    break;
 
-	case 'c':	/* command line contains source identifiers */
-	    flags |= PMSERIES_OPT_SOURCE;
+	case 'c':	/* path to .ini configuration file */
+	    inifile = opts.optarg;
 	    break;
 
 	case 'd':	/* command line contains series identifiers */
@@ -813,8 +1089,12 @@ main(int argc, char *argv[])
 	    flags |= PMSERIES_FAST;
 	    break;
 
-        case 'h':
-	    hostname = opts.optarg;
+	case 'g':
+	    match = sdsnew(opts.optarg);
+	    break;
+
+	case 'h':	/* Redis host to connect to */
+	    redis_host = opts.optarg;
 	    break;
 
 	case 'i':	/* command line contains series identifiers */
@@ -843,11 +1123,11 @@ main(int argc, char *argv[])
 	    break;
 
 	case 'n':	/* report label names only, not values */
-	    flags |= PMSERIES_ONLY_NAMES;
+	    flags |= (PMSERIES_OPT_LABELS|PMSERIES_ONLY_NAMES);
 	    break;
 
-        case 'p':	/* Redis port to connect to */
-	    port = (unsigned int)strtol(opts.optarg, NULL, 10);
+	case 'p':	/* Redis port to connect to */
+	    redis_port = (unsigned int)strtol(opts.optarg, NULL, 10);
 	    break;
 
 	case 'q':	/* command line contains query string */
@@ -855,20 +1135,45 @@ main(int argc, char *argv[])
 	    split = space;
 	    break;
 
-	case 'S':	/* report source identifiers, ala pminfo -S */
-	    flags |= PMSERIES_SOURCEID;
-	    break;
-
 	case 's':	/* report series identifiers, ala pminfo -s */
 	    flags |= PMSERIES_SERIESID;
 	    break;
 
+	case 'S':	/* command line contains source identifiers */
+	    flags |= PMSERIES_OPT_SOURCE;
+	    break;
+
 	default:
 	    opts.errors++;
 	    break;
 	}
     }
 
+    /*
+     * Parse the configuration file, extracting a dictionary of key/value
+     * pairs.  Each key is "section.name" and values are always strings.
+     * If no config given, default is /etc/pcp/pmproxy.conf (in addition,
+     * local user path settings in $HOME/.pcp/pmproxy.conf are merged) -
+     * pmseries(1) uses only keys from the [pmseries] section, but share
+     * the main pmproxy.conf file(s) (via symlink) for user convenience.
+     */
+    if ((config = pmIniFileSetup(inifile)) == NULL) {
+	pmprintf("%s: cannot setup from configuration file %s\n",
+			pmGetProgname(), inifile? inifile : "pmseries.conf");
+	opts.errors++;
+    } else {
+	/*
+	 * Push command line options into the configuration, and ensure
+	 * we have some default for attemping Redis server connections.
+	 */
+	if ((option = pmIniFileLookup(config, "pmseries", "servers")) == NULL ||
+	    (redis_host != NULL || redis_port != 6379)) {
+	    option = sdscatfmt(sdsempty(), "%s:%u",
+			redis_host? redis_host : "localhost", redis_port);
+	    pmIniFileUpdate(config, "pmseries", "servers", option);
+	}
+    }
+
     if (flags & PMSERIES_OPT_ALL)
 	flags |= PMSERIES_META_OPTS;
 
@@ -878,12 +1183,12 @@ main(int argc, char *argv[])
 			pmGetProgname());
 	opts.errors++;
     }
-    if ((flags & PMSERIES_OPT_LOAD) && (flags & PMSERIES_OPT_QUERY)) {
+    else if ((flags & PMSERIES_OPT_LOAD) && (flags & PMSERIES_OPT_QUERY)) {
 	pmprintf("%s: error - cannot use load and querying options together\n",
 			pmGetProgname());
 	opts.errors++;
     }
-    if ((flags & PMSERIES_OPT_QUERY) &&
+    else if ((flags & PMSERIES_OPT_QUERY) &&
 	(flags & (PMSERIES_META_OPTS | PMSERIES_OPT_SOURCE))) {
 	pmprintf("%s: error - cannot use query and metadata options together\n",
 			pmGetProgname());
@@ -892,19 +1197,42 @@ main(int argc, char *argv[])
 
     if (flags & (PMSERIES_FULLPMID | PMSERIES_FULLINDOM))
 	flags |= PMSERIES_NEED_DESCS;
-    if (flags & (PMSERIES_SOURCEID | PMSERIES_SERIESID))
+    if (flags & PMSERIES_SERIESID)
 	flags |= PMSERIES_NEED_DESCS;
 
     if (flags & PMSERIES_OPT_LABELS)
 	flags |= PMSERIES_NEED_INSTS;
 
-    if (!(flags & (PMSERIES_META_OPTS|PMSERIES_OPT_LOAD|PMSERIES_OPT_SOURCE)))
-	if (!(flags & (PMSERIES_NEED_DESCS|PMSERIES_NEED_INSTS)))
-	    flags |= PMSERIES_OPT_QUERY;	/* default is to query */
+    /*
+     * Determine default mode if no specific options presented.
+     * If all parameters are series hashes, assume --all metadata
+     * mode otherwise assume its a --query request.
+     */
+    if (!(flags & (PMSERIES_META_OPTS|PMSERIES_OPT_LOAD|PMSERIES_OPT_SOURCE)) &&
+	!(flags & (PMSERIES_NEED_DESCS|PMSERIES_NEED_INSTS))) {
+	for (c = opts.optind; c < argc; c++) {
+	    if (strlen(argv[c]) != 40)
+		break;
+	}
+	if (c != argc || opts.optind == argc)
+	    flags |= PMSERIES_OPT_QUERY;
+	else
+	    flags |= PMSERIES_OPT_ALL | PMSERIES_META_OPTS | PMSERIES_SERIESID;
+    }
 
-    if (opts.optind == argc && (flags & PMSERIES_OPT_QUERY)) {
-	pmprintf("%s: error - no --query string provided\n", pmGetProgname());
-	opts.errors++;
+    if (opts.optind == argc && !opts.errors) {
+	if ((flags & PMSERIES_OPT_QUERY)) {
+	   pmprintf("%s: error - no query string provided\n",
+			   pmGetProgname());
+	   opts.errors++;
+	}
+	else if (!(flags & (PMSERIES_META_OPTS|PMSERIES_OPT_SOURCE)) ||
+		 /* --all needs a timeseries identifier to work on */
+		 ((flags & PMSERIES_OPT_ALL) && opts.optind == argc)) {
+	    pmprintf("%s: error - no series string(s) provided\n",
+			    pmGetProgname());
+	    opts.errors++;
+	}
     }
 
     if (opts.errors || (opts.flags & PM_OPTFLAG_EXIT)) {
@@ -916,22 +1244,31 @@ main(int argc, char *argv[])
     if (pmLogLevelIsTTY())
 	flags |= PMSERIES_COLOUR;
 
-    settings.hostspec = sdscatprintf(sdsempty(), "%s:%u", hostname, port);
-
     if (opts.optind == argc)
 	query = sdsempty();
     else
 	query = sdsjoin(&argv[opts.optind], argc - opts.optind, (char *)split);
 
-    if (flags & PMSERIES_OPT_LOAD)
-	sts = series_load(&settings, query, flags);
-    else if (flags & PMSERIES_OPT_QUERY)
-	sts = series_query(&settings, query, flags);
-    else if ((flags & PMSERIES_OPT_SOURCE) && !(flags & PMSERIES_META_OPTS))
-	sts = series_source(&settings, query, flags);
-    else
-	sts = series_report(&settings, query, flags);
+    dp = series_data_init(flags, query);
+    dp->loop = uv_default_loop();
+    dp->args.pattern = match;
 
-    sdsfree(query);
-    return sts;
+    dp->settings.callbacks.on_match = on_series_match;
+    dp->settings.callbacks.on_desc = on_series_desc;
+    dp->settings.callbacks.on_inst = on_series_inst;
+    dp->settings.callbacks.on_labelmap = on_series_labelmap;
+    dp->settings.callbacks.on_instance = on_series_instance;
+    dp->settings.callbacks.on_context = on_series_context;
+    dp->settings.callbacks.on_metric = on_series_metric;
+    dp->settings.callbacks.on_value = on_series_value;
+    dp->settings.callbacks.on_label = on_series_label;
+    dp->settings.callbacks.on_done = on_series_done;
+
+    dp->settings.module.on_info = on_series_info;
+    dp->settings.module.on_setup = on_series_setup;
+
+    pmSeriesSetEventLoop(&dp->settings.module, dp->loop);
+    pmSeriesSetConfiguration(&dp->settings.module, config);
+
+    return pmseries_execute(dp);
 }
